"""Exceptions triggered to warn about internal errors related to Bluesky DataParser Factories classes."""


class BlueskyDataParserFactoryError(Exception):
    """Custom Exception to warn when any error occurs inside Bluesky DataParser Factories."""



"""Exceptions triggered to warn about internal errors related to Assonant data logger classes."""


class AssonantDataLoggerError(Exception):
    """Custom Exception to warn when any error occurs inside AssonantDataLogger."""



from typing import Any, Dict, List

from epics import PV

from assonant.data_classes.enums import ValuePlaceholders


class PVCollector:
    """PV Collector.

    Object responsible to abstracting the collection of values from multiple EPICS PVs.
    """

    pv_handlers = {}

    def __init__(self, pv_names: List[str]):
        """PV Collector Constructor

        Args:
            pv_names (List[str]): List of PV names which the collector will be responsible for.
        """

        # Pre-start PV handler for all passed PVs as it is faster than doing
        # it only when acquisition is demanded.
        for pv_name in pv_names:
            if pv_name not in self.pv_handlers:
                self.pv_handlers[pv_name] = PV(pv_name, connection_timeout=0.1)

        # Guarantee connection for PVs
        for pv_name in self.pv_handlers.keys():
            self.pv_handlers[pv_name].wait_for_connection()

    def acquire_data(self) -> Dict[str, Any]:
        """Call each pv handler to acquire and return current PV value.

        Example of returned data structure:

        {
            PV1_NAME: ACQUIRED_VALUE,
            PV2_NAME: ACQUIRED_VALUE,
            ...
        }

        Returns:
            Dict: Dict with keys being the PV name and value the acquired value for that PV.
        """

        acquisition_result = {}

        # Get value from all PVs
        for pv_name in self.pv_handlers.keys():
            if self.pv_handlers[pv_name].wait_for_connection():
                acquisition_result[pv_name] = self.pv_handlers[pv_name].get()
            else:
                acquisition_result[pv_name] = ValuePlaceholders.PV_NOT_CONNECTED.value

        return acquisition_result



import json
from typing import Any, Dict, List, Union

from .data_parser import DocumentsDataParser
from .exceptions import BlueskyDataParserFactoryError


class DataParserFactory:
    """Bluesky DataParser Factory Class.

    The role of this class is to handle type checking and return the correct DataParser to deal
    with given Bluesky data source. It is also responsible to raise a exception if passed source
    is from a unsupported type.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Component objects.
    """

    @staticmethod
    def create_data_parser(
        bluesky_data_source: Union[List[List[Union[str, Dict[str, Any]]]], str]
    ) -> DocumentsDataParser:
        """Instatiates a Suitable Bluesky DataParser to handle data parsing tasks over passed data source.

        Bluesky data must always be structure as the Bluesky Event-Model.
        Reference to it: https://github.com/bluesky/event-model

        Args:
            bluesky_data_source (Union[List[List[Union[str, Dict[str, Any]]]], str]): DataSource containing
            the bluesky data. Currently, supported formats are: JSON file containing bluesky events or a
            List of bluesky documents data presented also as  a list. The list with the document data has
            2 positions, the 1st is the bluesky identification for the document
            (start, event, stop, descriptor, ...) and the 2nd position the dict containing the document
            data.
        Raises:
            BlueskyDataParserError: Failed during DataParser creation or unsupported bluesky_data_source type.

        Returns:
            DataParser: Specific Bluesky DataParser able to handle tasks over passed bluesky_data_source.
        """
        if type(bluesky_data_source) is str:
            try:
                with open(bluesky_data_source, "r") as json_file:
                    bluesky_documents_list = json.load(json_file)
            except Exception as e:
                raise BlueskyDataParserFactoryError(
                    f"Failed to load bluesky documents data from JSON file with path: {bluesky_data_source}"
                ) from e
            return DocumentsDataParser(bluesky_documents_list=bluesky_documents_list)
        elif type(bluesky_data_source) is list:
            return DocumentsDataParser(bluesky_documents_list=bluesky_data_source)
        else:
            raise BlueskyDataParserFactoryError(
                f"Bluesky data source of type {type(bluesky_data_source)} not supported!"
            )



"""Exceptions triggered to warn about internal errors related to Assonant metadata retriever classes."""


class AssonantMetadataRetrieverError(Exception):
    """Custom Exception to warn when any error occurs inside AssonantMetadataRetriever."""



"""Assonant Monochromator component data class."""

from .component import Component


class Monochromator(Component):
    """Data class to handle all data required to define a monochromator."""


class MonochromatorCrystal(Component):
    """Data class to handle all data required to define a crystal from a Monochromator."""


class MonochromatorVelocitySelector(Component):
    """Data class to handle all data required to define a Velocity Selector from a Monochromator."""



"""Assonant class to handle writing data from AssonantDataClasses into files."""

import os
from typing import Generator, List

from assonant.data_classes import AssonantDataClass

from ._assonant_file_writer_interface import IAssonantFileWriter
from ._file_writer_factory import FileWriterFactory
from .exceptions import FileAlreadyExistsError


class AssonantFileWriter(IAssonantFileWriter):
    """Assonant File Writer. Wrapper class to deal with writing data from AssonantDataClass into files.

    Wrapper class that abstracts all process related to creating specific file writers,
    writing data on files following the requirements of the given format.
    """

    _factory = FileWriterFactory()

    def __init__(self, file_format: str):
        # Persist file format
        self.file_format = file_format

        # Create specific file writer based on chosen file format
        self.file_writer = self._factory.create_file_writer(file_format)
        self.file_extension = self.file_writer.get_file_extension()

    def write_data(self, filepath: str, filename: str, data: AssonantDataClass):
        """Method for writing data using the specific FileWriter respective to define file format.

        If fail to save on target filepath, in order to avoid losing acquired data, it will try
        to save locally in current working directory.

        Args:
            filepath (str): Path where file will be saved.
            filename (str): Name that will be given to file.
            data (AssonantDataClass): AssonantDataClass that contains data to be saved in file.
        """
        try:
            # Create target directory if it does not exist
            os.makedirs(filepath, exist_ok=True)

            indexed_filename_generator = self._indexed_name_generator(filename)
            indexed_filename = filename

            while True:
                try:
                    self.file_writer.write_data(filepath, indexed_filename, data)
                    break  # If no exception was raised, data was sucessfully written.
                except FileAlreadyExistsError:
                    # In case the file already exists, test next indexed_filename and try writing data again
                    indexed_filename = next(indexed_filename_generator)
                    continue
                except Exception as e:
                    # Any other kind of Exception signalizes a critical failure!
                    raise e
        except Exception as e:
            # Treat Critical Failure
            workdir_filepath = os.path.join(os.getcwd())
            print(
                f"Failed to save data at {filepath} due the following exception: {repr(e)}. Trying to save file in current workdir..."
            )
            try:
                unique_filename = self._create_unique_filename(
                    workdir_filepath, filename
                )  # Create unique filename to avoid conflict failure.
                self.file_writer.write_data(workdir_filepath, unique_filename, data)
                print(f"File saved at current work directory (path: {workdir_filepath})")
            except Exception as e:
                print("Failed to save data locally in current work directory. Data lost.")
                raise e

    def get_file_extension(self) -> str:
        """Getter method for the file extension of currently handled file writer.

        Returns:
            str: File extenstion of urrently handled file writer.
        """
        return self.file_extension

    def get_supported_file_formats(self) -> List[str]:
        """Getter method for current supported file formats.

        Returns:
            List[str]: List containing curretly supported file formats.
        """
        return self._factory.get_supported_file_formats()

    def _create_unique_filename(self, filepath: str, filename: str) -> str:
        """Return a unique file name to be saved on target filepath.

        Args:
            filepath (str): Path where file will be saved
            filename (str): Current filename.

        Returns:
            str: Unique filename that can be saved on target filepath. If current filename
            is already unique, it will be returned as it is. Otherwise, a index will be
            concatenated to the filename
        """
        new_name_generator = self._indexed_name_generator(filename)
        filename = next(new_name_generator)
        complete_file_path = self._mount_complete_file_path(filepath, filename)
        while os.path.exists(complete_file_path):
            filename = next(new_name_generator)
            complete_file_path = self._mount_complete_file_path(filepath, filename)
        return filename

    def _mount_complete_file_path(self, filepath: str, filename: str) -> str:
        """Mount complete path to the file.

        Args:
            filepath (str): Path where file will be saved.
            filename (str): Name of the file.

        Returns:
            str: Complete path for the file: File path + file name + file extension
        """
        return os.path.join(filepath, os.path.splitext(filename)[0]) + self.file_extension

    def _indexed_name_generator(self, name: str) -> Generator[str, None, None]:
        """Generator for creating indexed names.

        Args:
            name (str): Reference name to be indexed by generator.

        Yields:
            Generator[str, None, None]: New indexed name based on passed name.
        """
        i = 1
        while True:
            new_name = f"{name}_{i:06d}"
            yield new_name
            i += 1



"""
from typing import Tuple

import pytest

from assonant.data_classes import Entry
from assonant.data_classes.data_handlers import Axis, DataField, DataHandler, TimeSeries
from assonant.data_classes.enums import TransformationType
from assonant.data_classes.factories import (
    AssonantComponentFactory,
    AssonantDataHandlerFactory,
    AssonantEntryFactory,
)
from assonant.file_writer import AssonantFileWriter
from assonant.naming_standards import BeamlineName


@pytest.fixture(scope="session")
def good_nexus_file_writer() -> AssonantFileWriter:
    return AssonantFileWriter(file_format='nexus')

@pytest.fixture(scope="session")
def int_data_field_complete() -> Tuple[str,DataField]:
    return (
        'complete_int_data_field',
        AssonantDataHandlerFactory.create_data_field(
            value=10,
            unit='mm',
            extra_metadata={
                'metadata_1': 'I am a field metadata',
                'metadata_2': 12.34
            }
        )
    )

def int_data_field_missing_unit() -> Tuple[str,DataField]:
    return (
        'complete_int_data_field',
        AssonantDataHandlerFactory.create_data_field(
            value=10,
            extra_metadata={
                'metadata_1': 'I am a field metadata',
                'metadata_2': 12.34
            }
        )
    )


@pytest.fixture(scope="session")
def example_entry() -> Entry:

    mirror = AssonantComponentFactory.create_component_by_class_name('Mirror','M1')
    print(mirror)
    mirror.create_and_add_position(
        name='rx',
        transformation_type=TransformationType.ROTATION,
        value='65',
        unit='degrees'
    )
    mirror.create_and_add_position(
        name='ry',
        transformation_type=TransformationType.ROTATION,
        value='30',
        unit='degrees'
    )

    entry = AssonantEntryFactory.create_entry('entry', BeamlineName.MANACA)
    print(entry)
    return entry.beamline.add_subcomponent([mirror])
    
@pytest.fixture(scope="function")
def base_tmp_dir_path(tmp_path_factory) -> str:
    return str(tmp_path_factory.mktemp('test'))


@pytest.fixture(scope="function")
def test_write_data_fields():




#def test_nexus_write(good_nexus_file_writer: AssonantFileWriter, base_tmp_dir_path: str, example_entry: Entry):
#
#    good_nexus_file_writer.write_data(
#        base_tmp_dir_path,
#        'test_filename',
#        example_entry
#    )
#
#    assert True==True




#def test_none_field_write():

#def test_none_name_field_write():

#def test_none_name_group_write():

#def test_subcomponent_write():

#def test_nexus_entries_write():

#def test_nexus_entry_conflict_handling():

#def test_local_writing_after_failure():

#def test_failure_due_to_wrong_filepath():

#def test_failure_due_to_invalid_format():
"""



"""Assonant Data Logging module.

Assonant toolkit for automatically logging data based on data mapping/configration files.
"""

from .assonant_data_logger import AssonantDataLogger

__all__ = [
    "AssonantDataLogger",
]



from pathlib import Path

from ._assonant_metadata_retriever_interface import IAssonantMetadataRetriever
from ._csv_metadata_retriever import CSVMetadataRetriever
from .enums import MetadataSourceFileFormat
from .exceptions import AssonantMetadataRetrieverError


class MetadataRetrieverFactory:
    """Metadata Retriever Factory.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Metadata Retrievers.
    """

    def create_metadata_retriever(self, metadata_source_file_path: str) -> IAssonantMetadataRetriever:
        """Public method that abstracts metadata retriever creation process for the factory user.

        Internally, this method deals with validation and specific Metadata Retriever
        creation

        Args:
            metadata_source_file_path (str): Path to metadata source file.

        Raises:
            AssonantMetadataRetrieverError: An error occured during the creation of
            the respective Metadata Retriever.

        Returns:
            IAssonantMetadataRetriever: Data Retriever instance which implements the
            IAssonantMetadataRetriever interface for the given data_source_file_format.
        """
        file_format = self._get_file_format(metadata_source_file_path)

        self._validate_file_format(file_format)

        if file_format == MetadataSourceFileFormat.CSV.value:
            return self._create_csv_metadata_retriever(metadata_source_file_path)

        raise AssonantMetadataRetrieverError(
            f"'{file_format}' file format is set as supported but its creation method is not implemented."
        )

    def _create_csv_metadata_retriever(self, metadata_source_file_path: str) -> CSVMetadataRetriever:
        """Create CSVMetadataRetriver

        Args:
            metadata_source_file_path (str): Path to metadata source file

        Returns:
            CSVDataRetriever: Instance of CSVMetadataRetriever to deal data from passed data_source_file_path.
        """
        return CSVMetadataRetriever(csv_file_path=metadata_source_file_path)

    def _validate_file_format(self, file_format: str):
        """Check if file format is supported

        Args:
            file_format (str): Data source file format extension.

        Raises:
            AssonantDataRetrieverError: File format is not supported
        """

        if file_format not in MetadataSourceFileFormat._value2member_map_:
            raise AssonantMetadataRetrieverError(f"'{file_format}' is not supported by Metadata Retriever!")

    def _get_file_format(self, data_source_file_path: str) -> str:
        """Get file extension from given path

        Args:
            data_source_file_path (str): Path to data source file.

        Raises:
            AssonantDataRetrieverError: Erro raised if path passed is not from a file.

        Returns:
            str: File extesion without '.' character.
        """

        with Path(data_source_file_path) as path:
            if path.is_file():
                # Get file extension and remove the '.' character
                return path.suffix[1:]
            else:
                raise AssonantMetadataRetrieverError(f"'{data_source_file_path}' is not a file!")



"""Assonant data logger enums.

This submodule defines Enumerations classes used to standardize options related to data logging.
"""

from .value_placeholders import ValuePlaceholders

__all__ = ["CSVColumn", "ConfigFileFormat", "ValuePlaceholders"]



from abc import ABC, abstractmethod
from typing import Any


class IAssonantDataSender(ABC):
    """Assonant Data Sender Interface.

    This interface defines what must be implemented by a class to be a data
    sender in Assonant environment
    """

    @abstractmethod
    def send_data(self, payload: Any):
        """Method resposible for sending a payload through Data Sender specific communication method.

        Args:
            payload (Any): Any data that will be send. Type verification and specific treatment are responsibility
            from each specific DataSender.
        """



import os

from nexusformat.nexus import NeXusError, nxopen

from assonant._nexus import NeXusObjectFactory
from assonant.data_classes import AssonantDataClass

from ._assonant_file_writer_interface import IAssonantFileWriter
from .exceptions import (
    AssonantFileWriterError,
    FileAlreadyExistsError,
    NeXusObjectAlreadyExistsError,
)


class NexusFileWriter(IAssonantFileWriter):
    """NeXus File Writer.

    File Writer specialized on accessing data from AssonantDataClasses and
    write into a HDF5 file following NeXus data model standards.
    """

    def __init__(self):
        self.nexus_factory = NeXusObjectFactory()
        self.file_extension = ".nxs"

    def write_data(self, filepath: str, filename: str, data: AssonantDataClass):
        """Method for writing data into a NeXus file.

        File will be created if it doesn't exist yet, otherwise data will be add to NXroot object of the NeXus file.

        Args:
            filepath (str): Path where NeXus file will be saved.
            filename (str): Name that will be given to file.
            data (AssonantDataClass): AssonantDataClass that contains data
            to be saved in file.
        """
        nexus_filepath = os.path.join(filepath, os.path.splitext(filename)[0]) + self.file_extension

        mode = "rw" if os.path.isfile(nexus_filepath) else "a"

        try:
            with nxopen(nexus_filepath, mode=mode) as nxf:
                nxobject = self.nexus_factory.create_nxobject(data)
                try:
                    nxf.insert(nxobject)
                except NeXusError:
                    # insert method returns NeXusError when NXobject already exist in file
                    raise NeXusObjectAlreadyExistsError(
                        f"NeXus object {nxobject} already exisits on NeXus file from {nexus_filepath}"
                    )
        except NeXusObjectAlreadyExistsError as e:
            # To the upper layer, this must be faced as a FileAlreadyExistError, so raise this Exception.
            raise FileAlreadyExistsError(
                "Due to conflict on writing NeXus object, file is considered to be already existent!"
            ) from e
        except Exception as e:
            raise AssonantFileWriterError(f"NeXusFileWriter failed to write data into {filepath}") from e

    def get_file_extension(self) -> str:
        """Getter method for the file extension of currently handled file writer.

        Returns:
            str: File extenstion of urrently handled file writer.
        """
        return self.file_extension



"""Assonant Beam Visualiation System (BVS) component data class."""
from .component import Component


class BVS(Component):
    """Data class to handle all data required to define a Beam Visualization System (BVS)."""



"""Exceptions triggered to warn about internal errors related to Bluesky DataParser class."""


class BlueskyDataParserError(Exception):
    """Custom Exception to warn when any error occurs inside Bluesky DataParser."""



"""Exceptions triggered to warn about internal errors related to data classes Factories classes."""


class AssonantDataHandlerFactoryError(Exception):
    """Custom Exception to warn when any error occurs inside AssonantDataHandlerFactory."""


class AssonantComponentFactoryError(Exception):
    """Custom Exception to warn when any error occurs inside AssonantComponentFactory."""


class AssonantEntryFactoryError(Exception):
    """Custom Exception to warn when any error occurs inside AssonantEntryFactory."""



"""Assonant data logger exceptions.

This submodule defines Exceptions used all over the data logger module and its submodules.
"""

from .data_loggers_exceptions import AssonantDataLoggerError

__all__ = ["AssonantDataLoggerError"]



"""Assonant Experiment Done Event class."""

from .assonant_event import AssonantEvent


class ExperimentDone(AssonantEvent):
    """Event class to signalize that a experiment has finished.

    This event class gather data required to signalize that a experiment has finished and also
    that allow accessing the data generated during the experiment.
    """

    beamline: str
    filepath: str



from typing import Any, Dict, List

from assonant.naming_standards import AcquisitionMoment

from ._assonant_metadata_retriever_interface import IAssonantMetadataRetriever
from ._metadata_retriever_factory import MetadataRetrieverFactory
from .exceptions.metadata_retrievers_exceptions import AssonantMetadataRetrieverError


class AssonantMetadataRetriever(IAssonantMetadataRetriever):
    """Assonant Metadata Retriever.

    Wrapper class that abstracts all process related to creating specific metadata retrievers
    for specific file formats.
    """

    _factory = MetadataRetrieverFactory()

    def __init__(self, metadata_source_file_path: str):

        # Persist passed data_source_file_path
        self.metadata_source_file_path = metadata_source_file_path

        self.metadata_retriever = self._factory.create_metadata_retriever(metadata_source_file_path)

    def get_pv_names_by_acquisition_moment(self, acquisition_moment: AcquisitionMoment) -> List[str]:
        """Return a List with all PV names related to the passed AcquisitionMoment.

        Example of returned structure: [PV1_NAME, PV2_NAME, ...]

        Args:
            acquisition_moment (AcquisitionMoment): Target Acquisition Moment used for selecting which
            PV names will be returned.

        Raises:
            AssonantMetadataRetrieverError: Failure to retrieve requested data.

        Returns:
            List[str]: List containing all PV names related to passed AcquisitionMoment.
        """
        try:
            return self.metadata_retriever.get_pv_names_by_acquisition_moment(acquisition_moment)
        except Exception as e:
            raise AssonantMetadataRetrieverError(
                f"Failed to retrieve PV names from {self.metadata_source_file_path} for '{acquisition_moment}' Acquisition Moment"
            ) from e

    def get_pvs_info(self, pv_names: List[str]) -> Dict[str, Dict[str, Any]]:
        """Return all information about the PV paired by the its name.

        Example of return structure.:
        {
            PV_1_NAME_: {
                component_info:
                {
                    name: COMPONENT_NAME,
                    class: COMPONENT_CLASS
                    subcomponent_of: ASCENDANT_COMPONENT_NAME (if exists)
                },
                data_handler_info:
                {
                    name: FIELD_NAME,
                    value: PLACEHOLDER_VALUE
                    unit: FIELD_UNIT (if exists),
                    transformation_type: FIELD_TRANSFORMATION_TYPE (if exists),
                    PV_1_EXTRA_METADATA_1_NAME: PV_1_EXTRA_METADATA_1_VALUE (if exists),
                    PV_1_EXTRA_METADATA_2_NAME: PV_1_EXTRA_METADATA_2_VALUE (if exists),
                    ...
                }
            },
            PV_2_NAME_: {
                component_info:
                {
                    name: COMPONENT_NAME,
                    class: COMPONENT_CLASS
                    subcomponent_of: ASCENDANT_COMPONENT_NAME (if exists)
                },
                data_handler_info:
                {
                    name: FIELD_NAME,
                    value: PLACEHOLDER_VALUE
                    unit: FIELD_UNIT (if exists),
                    transformation_type: FIELD_TRANSFORMATION_TYPE (if exists),
                    PV_2_EXTRA_METADATA_1_NAME: PV_2_EXTRA_METADATA_1_VALUE (if exists),
                    PV_2_EXTRA_METADATA_2_NAME: PV_2_EXTRA_METADATA_2_VALUE (if exists),
                    ...
                }
            },
        }

        Args:
            pv_names (List[str]): List with PV names which data field info will be fetched.

        Raises:
            AssonantMetadataRetrieverError: Failure to retrieve requested data.

        Returns:
            Dict[str, Dict[str, Any]]: Dictionary containing all PV info paired by the PV name.
        """
        try:
            return self.metadata_retriever.get_pvs_info(pv_names)
        except Exception as e:
            raise AssonantMetadataRetrieverError(
                f"Failed to retrieve pvs info from {self.metadata_source_file_path}"
            ) from e

    def get_subcomponents_mapping(self) -> Dict[str, List[str]]:
        """Return a Dictionary mapping component names with the name of its subcomponents.

        This call MUST be indepedent of the AcquisitionMoment as components hierarchy is not related to
        acquisition directly but logical organization. If AcquisitionMoment is considered here, it may
        generate hierarchy inconsistency due to the possibility of parent Component exist in some
        AcquisitionMoments and not in others.

        Important considerations about the returned structure:
            * The returned structure contains only Components that have at least one Subcomponent.
            * Subcomponent can contain Subcomponents, that said, names that appears as Subcomponents
            for a Component may also appear as Component with other Subcomponents. Treating how to
            deal with this is responsibility of who is calling this method.

        Example of returned structure:
        {
            COMPONENT_1_NAME: [SUBCOMPONENT_1_NAME, SUBCOMPONENT_2_NAME, ...],
            COMPONENT_2_NAME: [SUBCOMPONENT_1_NAME, SUBCOMPONENT_2_NAME, ...],
            ...
        }

        Raises:
            AssonantMetadataRetrieverError: Failure to retrieve requested data.

        Returns:
            Dict[str, List[str]]: Dictionary mapping the name of each Component that contains at
            least one Subcomponent with their respective List of Subcomponents names.
        """
        try:
            return self.metadata_retriever.get_subcomponents_mapping()
        except Exception as e:
            raise AssonantMetadataRetrieverError(
                f"Failed to retrieve subcomponents mapping from {self.metadata_source_file_path}"
            ) from e

    def get_component_info(self, component_name: str) -> Dict[str, Dict[str, Any]]:
        """Retrieve info about a specific Component based on passed component_name.

        Example of returned structure:
            CHECK get_pvs_info() 'component_info' structure. IT MUST BE THE SAME.

        Args:
            component_name (str): Component name to retrieve information about.

        Raises:
            AssonantMetadataRetrieverError: Failure to retrieve requested data.

        Returns:
            Dict[str, Dict[str, Any]]: Component info dictionary following the same subdictionary
            returned by get_pvs_info() method to represent component info.
        """
        try:
            return self.metadata_retriever.get_component_info(component_name)
        except Exception as e:
            raise AssonantMetadataRetrieverError(
                f"Failed to retrieve info of component named: '{component_name}'"
            ) from e



from typing import Any, Dict, List, Optional, Union

from assonant.data_classes import (
    AssonantComponentFactory,
    AssonantDataHandlerFactory,
    AssonantEntryFactory,
    Entry,
)
from assonant.data_classes.components import Beamline, Component
from assonant.data_classes.data_handlers import Axis, DataHandler
from assonant.data_classes.exceptions import AxisInsertionError, FieldInsertionError
from assonant.data_classes.hierarchizer import Hierarchizer
from assonant.data_retriever import AssonantDataRetriever
from assonant.file_writer import AssonantFileWriter
from assonant.metadata_retriever import AssonantMetadataRetriever
from assonant.naming_standards import AcquisitionMoment, BeamlineName, ExperimentStage
from assonant.path_builder import BeamlinePathBuilder

from ._pv_collector import PVCollector
from .exceptions import AssonantDataLoggerError


class AssonantDataLogger:
    """Assonant Data Logger.

    Object responsible to automate data collection, standardization and logging based on
    pré-defined configurations.
    """

    _supported_acquisition_moments_for_auto_collection = [AcquisitionMoment.START, AcquisitionMoment.END]

    def __init__(
        self,
        beamline_name: BeamlineName,
        pre_init_auto_collector_pv_connections: bool = False,
        data_schema_file_path: Optional[str] = None,
    ):
        """AssonantDataLogger Constructor.

        Args:
            beamline_name (BeamlineName): Beamline name where the experiment is being executed.
            pre_init_auto_collector_pv_connections (bool): Flag to control if connections to auto collectable PVs should
            be pre-initiallized during AssonantDataLogger instantiation or not. Defaults to False.
            data_schema_file_path (str): Path to data schema file.
        """

        self.beamline_name = beamline_name

        if data_schema_file_path is None:
            # No data schema file path passed? - Build it based on beamline name
            self.data_schema_file_path = self._auto_build_data_schema_file_path(beamline_name=beamline_name)
        else:
            # Use passed data schema file path
            self.data_schema_file_path = data_schema_file_path

        self.metadata_retriever = AssonantMetadataRetriever(metadata_source_file_path=self.data_schema_file_path)
        self.hierarchizer = Hierarchizer(self.data_schema_file_path, self.metadata_retriever)
        self.pv_names = {}
        self.pv_collectors = {}
        self.file_writer = AssonantFileWriter("nexus")
        self.path_builder = BeamlinePathBuilder()
        self.pv_collectors_innitialized = False
        if pre_init_auto_collector_pv_connections is True:
            self.init_auto_collector()

    def _auto_build_data_schema_file_path(self, beamline_name: BeamlineName) -> str:
        """Build path within storage to schema file automatically based on beamline_name.

        Args:
            beamline_name (BeamlineName): Enum containing the respective Beamline Name.

        Returns:
            str: Path within storage where schema file is stored.
        """
        self.path_builder.set_beamline_name(beamline_name.value.lower())
        return self.path_builder.build_path_to_data_schema()

    def _create_data_handler_from_pv_info(self, pv_info: Dict[str, Dict[str, Any]]) -> DataHandler:
        """Create data handler based on passed pv_info Dict.

        Args:
            pv_info (Dict[str, Dict[str, Any]]): Dict containing acquired PV info. The Dict follows must follow
            the structure proposed for the get_pvs_info() call from the IAssonantDataRetriever Interface.

        Returns:
            DataHandler: Specific DataHandler related to the passed info.
        """
        param_names = ["value", "unit", "transformation_type", "timestamps", "timestamps_unit"]
        ignored = ["name"]
        extra_metadata = {}
        params = {}

        # Search into dict which names are parameters for the create method from factory, the rest
        # is considered as extra_metadata
        data_handler_info = pv_info["data_handler_info"]
        for info_name in data_handler_info:
            if info_name not in ignored:
                if info_name not in param_names:
                    extra_metadata[info_name] = data_handler_info[info_name]
                else:
                    params[info_name] = data_handler_info[info_name]

        params["extra_metadata"] = extra_metadata

        return AssonantDataHandlerFactory.create_data_handler(**params)

    def _add_fields_to_component(self, component: Component, fields: Dict[str, DataHandler]):
        """Add fields from dict to passed Component.

        Args:
            component (Component): Component which fields will be inserted to.
            fields (dict[str, DataHandler]): Dict of fields that will be inserted on component.
        """
        for field_name in fields:
            try:
                component.add_field(name=field_name, new_field=fields[field_name])
            except FieldInsertionError:
                # To avoid breaking and not logging other data, errors in insertion are ignored and logged
                # TODO: Put this into a log file instead of printing it
                print(
                    f"{field_name} field was not add to {component.name} due to an Error during its insertion on fields dict"
                )

    def _add_positions_to_component(self, component: Component, positions: Dict[str, Axis]):
        """Add axis from dict to passed Component.

        Args:
            component (Component): Component which axis will be inserted to.
            positions (dict[str, Axis]): Dict of axis that will be inserted on component.
        """
        for axis_name in positions:
            try:
                component.add_position(name=axis_name, new_axis=positions[axis_name])
            except AxisInsertionError:
                # To avoid breaking and not logging other data, errors in insertion are ignored and logged
                # TODO: Put this into a log file instead of printing it
                print(
                    f"{axis_name} Axis was not add to {component.name} due to an Error during its insertion on positions dict"
                )

    def _create_component_from_pv_info(self, pv_info: Dict[str, Dict[str, Any]]) -> Component:
        """Create an Assonant Component based on passed dict with PV data.

        Args:
            pv_info (Dict[str, Dict[str, Any]]): Dict containing Component info retrieved from pv_info dict.
            The passed dict must follow the structure defined on get_pvs_info() method from IAssonantDataRetriever.

        Returns:
            Component: Assonant Component respective to passed PV data.
        """
        component_info = pv_info["component_info"]
        return AssonantComponentFactory.create_component_by_class_name(
            class_name=component_info["class"], component_name=component_info["name"]
        )

    def _wrap_components_into_entry(self, components: List[Component], entry: Entry) -> Entry:
        """Wrap components into a Entry object based on the AcquisitionMoment.

        Args:
            components (List[Component]): List of Components to wrap into the Entry object.
            entry (Entry): Entry which Components will be add to.

        Returns:
            Entry: Entry containing all passed components within a Beamline object.
        """
        for component in components:

            if isinstance(component, Entry):
                # Special case: Metadata related to Entry must be added to pre-created instance
                entry.take_data_handlers_from(entry=component)
            elif isinstance(component, Beamline):
                # Special case: Metadata related to beamline must be added to pre-created instance
                entry.beamline.take_data_handlers_from(component=component)
            else:
                entry.beamline.add_subcomponent(component)

        return entry

    def init_auto_collector(self):
        """Pre-establish PV collectors connections for auto collection functionality.

        Recommended to avoid having to establish connection with all PVs during a collect call.
        """

        acquisition_moments = [AcquisitionMoment.START, AcquisitionMoment.END]

        for acquisition_moment in acquisition_moments:

            # Try to retrieve PV names that will be collect on Acquisition Moment
            try:
                self.pv_names[acquisition_moment.value] = self.metadata_retriever.get_pv_names_by_acquisition_moment(
                    acquisition_moment
                )
            except Exception as e:
                raise AssonantDataLoggerError(
                    f"Failed to retrieve PV names for '{acquisition_moment.value}' Acquisition Moment!"
                ) from e

            # Try to initialize connection retrieve PVs
            try:
                self.pv_collectors[acquisition_moment.value] = PVCollector(self.pv_names[acquisition_moment.value])
            except Exception as e:
                raise AssonantDataLoggerError(
                    f"Faield to initialize PVCollector for PVs from '{acquisition_moment.value}' Acquisition Moment!"
                ) from e

        # Update PV collector status flag
        self.pv_collectors_innitialized = True

    def collect_and_log_pv_data(
        self,
        log_file_path: str,
        log_file_name: str,
        experiment_stage: ExperimentStage,
        acquisition_moment: AcquisitionMoment,
    ):
        """Trigger data logger to collect PV data based on AcquisitionMoment, standardize it and log it.

        Args:
            log_file_path (str): Path where log file will be saved.
            log_file_name (str): Name of the log file that will be saved on log_file_path.
            acquisition_moment (AcquisitionMoment): AcquisitionMoment which PV data will be collected and logged.

        """
        if acquisition_moment in self._supported_acquisition_moments_for_auto_collection:
            if self.pv_collectors_innitialized is False:
                self.init_auto_collector()
            acquired_pv_data = self.pv_collectors[acquisition_moment.value].acquire_data()
            self.log_collected_pv_data(
                log_file_path=log_file_path,
                log_file_name=log_file_name,
                experiment_stage=experiment_stage,
                acquisition_moment=acquisition_moment,
                pv_data=acquired_pv_data,
            )
        else:
            raise AssonantDataLoggerError(
                f"Experimental stage '{acquisition_moment.value}' is not supported yet for auto collection! Use the log_collected_data method instead!"
            )

    def log_collected_pv_data(
        self,
        log_file_path: str,
        log_file_name: str,
        experiment_stage: ExperimentStage,
        acquisition_moment: AcquisitionMoment,
        pv_data: Dict[str, Any],
    ):
        """Standardize and log collected PV data represented as {PV_NAME: COLLECTED_VALUE} Dict.

        Args:
            log_file_path (str): Path where log file will be saved.
            log_file_name (str): Name of the log file that will be saved on log_file_path.
            acquisition_moment (AcquisitionMoment): Current AcquisitionMoment which data was collected.
            pv_data (Dict[str, Any]): Collected PV data. The Dict structure must follow the same data structure
            and representation as the PVCollector acquire_data() method.
        """
        pv_names = [pv_name for pv_name in pv_data.keys()]
        pv_name_and_info_mapping = self.metadata_retriever.get_pvs_info(pv_names=pv_names)
        components = {}

        # Create DataHandlers for each acquired PV
        for pv_name, pv_info in pv_name_and_info_mapping.items():

            # Create DataHandler for PV field
            data_handler = self._create_data_handler_from_pv_info(pv_info)
            data_handler.set_value(pv_data[pv_name])
            data_handler_name = pv_info["data_handler_info"]["name"]

            component_name = pv_info["component_info"]["name"]

            # Create Component if doesn't already exist
            if component_name not in components.keys():
                components[component_name] = self._create_component_from_pv_info(pv_info)

            components[component_name].add_data_handler(name=data_handler_name, new_data_handler=data_handler)

        hierarchized_components = self.hierarchizer.hierarchize_components(list(components.values()))

        entry_name = "_".join([experiment_stage.value, acquisition_moment.value])
        entry = AssonantEntryFactory.create_entry(entry_name, self.beamline_name)
        entry = self._wrap_components_into_entry(hierarchized_components, entry)

        self.file_writer.write_data(log_file_path, log_file_name, entry)

    def log_collected_component_data(
        self,
        log_file_path: str,
        log_file_name: str,
        experiment_stage: ExperimentStage,
        acquisition_moment: AcquisitionMoment,
        components: Union[Component, List[Component]],
    ):
        """Standardize and log collected PV data already wrapped as Assonant Components.

        Args:
            log_file_path (str): Path where log file will be saved.
            log_file_name (str): Name of the log file that will be saved on log_file_path.
            acquisition_moment (AcquisitionMoment): Current AcquisitionMoment which data was collected.
            components (Union[Component, List[Component]]): List of Assonant Component objects with data to be logged.
        """
        print(
            "WARNING: LOG_COLLECTED_COMPONENT_DATA METHOD IS DISCOURAGED DUE TO THE LACK OF AUTOMATION AND"
            + "LACK OF COMPONENTS HIERARCHY STANDARDIZATION. ONLY USE THIS IF YOU ARE REALLY SURE"
            + "IN WHAT YOU ARE DOING!!!"
        )
        if isinstance(components, Component):
            components = [components]

        entry_name = "_".join([experiment_stage.value, acquisition_moment.value])
        entry = AssonantEntryFactory.create_entry(entry_name, self.beamline_name)
        entry = self._wrap_components_into_entry(components, entry)

        self.file_writer.write_data(log_file_path, log_file_name, entry)

    def log_bluesky_data_from_json_file(
        self,
        log_file_path: str,
        log_file_name: str,
        experiment_stage: ExperimentStage,
        acquisition_moment: AcquisitionMoment,
        json_file_path: str,
    ):
        """Log bluesky data on json files into a standardize way

        Args:
            log_file_path (str): Path where log file will be saved.
            log_file_name (str): Name of the log file that will be saved on log_file_path.
            experiment_stage (ExperimentStage): Enum related to the ExperimentStage in which the data was collected.
            acquisition_moment (AcquisitionMoment): Enum related to the AcquisitionMoment that data was collected
            within the ExperimentStage.
            json_file_path (str): Path to the json file containing the bluesky data.
        """

        # TODO: POSSIBLE ENHANCEMENT: FUTURALLY, DEPENDING ON PERFORMANCE REQUIREMENTS, SOME KIND OF CONTROL
        # IF INSTANTIATES A NEW BLUESKY DATA PARSER OR REUSE AN ALREADY EXISTING ONE MAY BE GOOD. THIS
        # WAY IT AVOIDS PROCESSING MULTUIPLE TIMES THE SAME DOCUMMENTS TO FIND THE SAME DATA.
        assonant_data_retriever = AssonantDataRetriever(data_source=json_file_path)

        pv_data = assonant_data_retriever.get_pv_data_by_acquisition_moment(acquisition_moment)
        # print(pv_data)
        self.log_collected_pv_data(log_file_path, log_file_name, experiment_stage, acquisition_moment, pv_data)

    def log_bluesky_data_from_documents_list(
        self,
        log_file_path: str,
        log_file_name: str,
        experiment_stage: ExperimentStage,
        acquisition_moment: AcquisitionMoment,
        documents_list: List[List[Union[str, Dict[str, Any]]]],
    ):
        """Log bluesky data passed as a dict of bluesky documments (which are also represented as dicts)

        Args:
            log_file_path (str): Path where log file will be saved.
            log_file_name (str): Name of the log file that will be saved on log_file_path.
            experiment_stage (ExperimentStage): Enum related to the ExperimentStage in which the data was collected.
            acquisition_moment (AcquisitionMoment): Enum related to the AcquisitionMoment that data was collected
            within the ExperimentStage.
            documents_list (List[List[Union[str, Dict[str, Any]]]]): List of bluesky documents data presented also as
            a list. The list with the document data has 2 positions, the 1st is the bluesky identification for the
            document (start, event, stop, descriptor, ...) and the 2nd position the dict containing the document
            data.
        """

        # TODO: POSSIBLE ENHANCEMENT: FUTURALLY, DEPENDING ON PERFORMANCE REQUIREMENTS, SOME KIND OF CONTROL
        # IF INSTANTIATES A NEW BLUESKY DATA PARSER OR REUSE AN ALREADY EXISTING ONE MAY BE GOOD. THIS
        # WAY IT AVOIDS PROCESSING MULTUIPLE TIMES THE SAME DOCUMMENTS TO FIND THE SAME DATA.
        assonant_data_retriever = AssonantDataRetriever(data_source=documents_list)

        pv_data = assonant_data_retriever.get_pv_data_by_acquisition_moment(acquisition_moment)
        # print(pv_data)
        self.log_collected_pv_data(log_file_path, log_file_name, experiment_stage, acquisition_moment, pv_data)



from pathlib import Path
from typing import Any

from ._assonant_data_retriever_interface import IAssonantDataRetriever
from ._bluesky_data_retriever import BlueskyDataRetriever
from .enums import DataSourceFileFormat
from .exceptions import AssonantDataRetrieverError


class DataRetrieverFactory:
    """Data Retriever Factory.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Data Retrievers.
    """

    def create_data_retriever(self, data_source: Any) -> IAssonantDataRetriever:
        """Public method that abstracts data retriever creation process for the factory user.

        Internally, this method deals with validation and specific Data Retriever
        creation

        Args:
            data_source (Any): Data source that data will be retrieved from.

        Raises:
            AssonantDataRetrieverError: An error occured during the creation of
            the respective Data Retriever.

        Returns:
            IAssonantDataRetriever: Data Retriever instance which implements the
            IAssonantDataRetriever interface for the given data_source_file_format.
        """

        if type(data_source) is str:
            # Data source is a file
            file_format = self._get_file_format(data_source)

            self._validate_file_format(file_format)

            if file_format == DataSourceFileFormat.JSON.value:
                # Data source is a JSON file with BlueskyData
                return self._create_bluesky_data_retriever(data_source)
            else:
                raise AssonantDataRetrieverError(f"File format {file_format} not supported!")
        elif type(data_source) is list:
            # Data source is an already loaded bluesky data into a dict of bluesky documents.
            return self._create_bluesky_data_retriever(data_source)
        else:
            raise AssonantDataRetrieverError("Data source not supported for data retrieving!")

    def _create_bluesky_data_retriever(self, data_source: Any) -> BlueskyDataRetriever:
        """Create BlueskyDataRetriver

        Args:
            data_source (str): Bluesky data source. It can be a JSON filepath or a List
            of lists containing the stream name and bluesky events data.

        Returns:
            BlueskyDataRetriever: Instance of BlueskyDataRetriver to deal data from passed data_source_file_path.
        """
        return BlueskyDataRetriever(bluesky_data_source=data_source)

    def _validate_file_format(self, file_format: str):
        """Check if file format is supported

        Args:
            file_format (str): Data source file format extension.

        Raises:
            AssonantDataRetrieverError: File format is not supported
        """

        if file_format not in DataSourceFileFormat._value2member_map_:
            raise AssonantDataRetrieverError(f"'{file_format}' is not supported by Data Retriever!")

    def _get_file_format(self, data_source_file_path: str) -> str:
        """Get file extension from given path

        Args:
            data_source_file_path (str): Path to data source file.

        Raises:
            AssonantDataRetrieverError: Erro raised if path passed is not from a file.

        Returns:
            str: File extesion without '.' character.
        """

        with Path(data_source_file_path) as path:
            if path.is_file():
                # Get file extension and remove the '.' character
                return path.suffix[1:]
            else:
                raise AssonantDataRetrieverError(f"'{data_source_file_path}' is not a file!")



"""Factories used to abstract and centralize Assonant data classes instanciation process."""

import importlib
from typing import Dict, List, Optional, Type, Union

import numpy as np

from assonant.naming_standards import BeamlineName

from .components import Beamline, Component
from .data_handlers import Axis, DataField, TimeSeries
from .entry import Entry
from .enums import TransformationType
from .exceptions import (
    AssonantComponentFactoryError,
    AssonantDataHandlerFactoryError,
    AssonantEntryFactoryError,
)


class AssonantComponentFactory:
    """Assonant Component Factory Class.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Component objects.
    """

    @staticmethod
    def create_component_by_class_name(class_name: str, component_name: str) -> Component:
        """Create an Assonant data class Component based on the class name passed in a string.

        Args:
            class_name (str): String with value equal to the name of a AssonantDataClass type.
            component_name (str): Name given to the component that will be created.

        Raises:
            AssonantComponentFactoryException: Passed 'class_name' argument is invalid.

        Returns:
            Component: Named Component object of type equals to the 'class_name' parameter.
        """
        try:
            component_class = getattr(importlib.import_module("assonant.data_classes.components"), class_name)
            component = component_class(name=component_name)
        except Exception as e:
            raise AssonantComponentFactoryError(
                f"An Error ocurred when creating Assonant Component by the following class name: {class_name}"
            ) from e
        return component

    @staticmethod
    def create_component_by_class_type(class_type: Component, component_name: str) -> Component:
        """Create an Assonant data class Component based on the class type passed.

        Args:
            class_type (Component): Assonant Component type of the class that will be created.
            component_name (str): Name given to the component that will be created.

        Raises:
            AssonantComponentFactoryException: Passed 'class_type' argument is invalid.

        Returns:
            Component: Named Component object of 'class_type' type.
        """
        try:
            component = class_type(name=component_name)
        except Exception as e:
            raise AssonantComponentFactoryError(
                f"An Error ocurred when creating Assonant Component by the following class type: {class_type}"
            ) from e
        return component


class AssonantDataHandlerFactory:
    """NeXus Data Handler Factory Class.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Data Handler objects.
    """

    @staticmethod
    def create_data_handler(
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        transformation_type: Optional[TransformationType] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps: Union[int, float, str, List, Type[np.ndarray], None] = None,
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ) -> Union[DataField, Axis, TimeSeries]:
        """Create the best fitting DataHandler object based on passed data.

        Args:
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter. Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]], optional): Dict
            containing any aditional metadata related to collected data. Defaults to {}.
            transformation_type (Optional[TransformationType], optional): TransformationType related to Axis data.
            Defaults to None.
            timestamps (Union[int, float, str, List, Type[np.ndarray], None], optional): Timestamps related to
            collected data. Defaults to None.
            timestamps_unit (Optional[str], optional): Measurement unit related to timestamps parameter.
            Defaults to None.
            timestamp_extra_metadata (Dict[str, Union[int, float, str, List, Type[np.ndarray], None]], optional):
            Dictionary containing extra metadata about timestamps field. Defaults to {}.

        Returns:
            Union[DataField, Axis, TimeSeries]: DataHandler object depeding on passed parameters:
            1) If timestamps != None AND transformationType != None method returns a TimeSeries with 'value'
            being an Axis.
            2) If timestamps != None AND transformationType == None method returns a  TimeSeries with 'value'
            being a DataField.
            3) If timestamps == None AND transformationType != None method returns an Axis.
            4) If timestamps == None AND transformationType == None method returns a DataField.

        """
        # Check key optional fields that changes the used DataHandler class
        if timestamps is None and timestamps_unit is None and timestamp_extra_metadata == {}:
            # It is not a TimeSeries monitored ...
            if transformation_type is not None:
                # Axis
                new_data_handler = AssonantDataHandlerFactory.create_axis(
                    transformation_type=transformation_type, value=value, unit=unit, extra_metadata=extra_metadata
                )
            else:
                # DataField
                new_data_handler = AssonantDataHandlerFactory.create_data_field(
                    value=value, unit=unit, extra_metadata=extra_metadata
                )
        else:
            # It is a TimeSeries monitored ...
            if transformation_type is not None:
                # Axis
                new_data_handler = AssonantDataHandlerFactory.create_timeseries_axis(
                    transformation_type=transformation_type,
                    value=value,
                    timestamps=timestamps,
                    unit=unit,
                    extra_metadata=extra_metadata,
                    timestamps_unit=timestamps_unit,
                    timestamp_extra_metadata=timestamp_extra_metadata,
                )
            else:
                # DataField
                new_data_handler = AssonantDataHandlerFactory.create_timeseries_field(
                    value=value,
                    timestamps=timestamps,
                    unit=unit,
                    extra_metadata=extra_metadata,
                    timestamps_unit=timestamps_unit,
                    timestamp_extra_metadata=timestamp_extra_metadata,
                )

        return new_data_handler

    @staticmethod
    def create_data_field(
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ) -> DataField:
        """Create new DataField handler.

        Args:
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]], optional): Dict
            containing any aditional metadata related to collected data. Defaults to {}.

        Raises:
            AssonantDataHandlerFactoryException: Failure during objects creation.

        Returns:
            DataField: New DataField object handlind given values.
        """
        try:
            new_field = DataField(
                value=value,
                unit=unit,
                extra_metadata=extra_metadata,
            )
        except Exception as e:
            raise AssonantDataHandlerFactoryError("Failed to create new DataField handler.") from e

        return new_field

    @staticmethod
    def create_axis(
        transformation_type: TransformationType,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ) -> Axis:
        """Create new Axis handler.

        Args:
            transformation_type (TransformationType): Type of transformation done by axis.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to axis
            collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing any aditional metadata related to collected data. Defaults to {}.

        Raises:
            AssonantDataHandlerFactory: Failure during objects creation.

        Returns:
            Axis: New Axis object handling given data.
        """
        try:
            new_axis = Axis(
                transformation_type=transformation_type,
                value=AssonantDataHandlerFactory.create_data_field(
                    value=value,
                    unit=unit,
                    extra_metadata=extra_metadata,
                ),
            )
        except Exception as e:
            raise AssonantDataHandlerFactoryError("Failed to create new Axis handler") from e

        return new_axis

    @staticmethod
    def create_timeseries_field(
        value: Union[int, float, str, List, Type[np.ndarray], None],
        timestamps: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ) -> TimeSeries:
        """Create new Timeseries handler to handle timestamped DataField data.

        Args:
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            timestamps (Union[int, float, str, List, Type[np.ndarray], None]): Timestamps related to data collected
            from the field.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing any aditional metadata related to collected data. Defaults to {}.
            timestamps_unit (Optional[str], optional): Measurement unit related to
            timestamp field. Defaults to None.
            timestamp_extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing extra metadata about timestamps field. Defaults to {}.

        Raises:
            AssonantDataHandlerFactoryException: Failure during objects creation.

        Returns:
            TimeSeries: New TimeSeries object handling given timestamped data.
        """
        try:
            new_field = TimeSeries(
                value=AssonantDataHandlerFactory.create_data_field(
                    value=value,
                    unit=unit,
                    extra_metadata=extra_metadata,
                ),
                timestamps=AssonantDataHandlerFactory.create_data_field(
                    value=timestamps,
                    unit=timestamps_unit,
                    extra_metadata=timestamp_extra_metadata,
                ),
            )
        except Exception as e:
            raise AssonantDataHandlerFactoryError("Failed to create new TimeSeries handler.") from e

        return new_field

    @staticmethod
    def create_timeseries_axis(
        transformation_type: TransformationType,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        timestamps: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ) -> Axis:
        """Create new TimeSeries handler to handle timestamped Axis data.

        Args:
            transformation_type (TransformationType): Transformation type of the Axis
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to
            axis collected data
            timestamps (Union[int, float, str, List, Type[np.ndarray], None]): Timestamps
            related to data collected from the axis.
            unit (Optional[str], optional): Measurement unit related to value
            field. Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing extra metadata about value field. Defaults to {}.
            tracked it as a TimeSeries. Defaults to None.
            timestamps_unit (Optional[str], optional): Measurement unit related to
            timestamp field. Defaults to None.
            timestamp_extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dict
            containing extra metadata about timestamps field. Defaults to {}.

        Raises:
            AssonantDataHandlerFactoryException: ailure during objects creation.

        Returns:
            Axis: New Axis object handling given timestamped data.
        """
        try:
            new_axis = Axis(
                transformation_type=transformation_type,
                value=AssonantDataHandlerFactory.create_timeseries_field(
                    value=value,
                    unit=unit,
                    extra_metadata=extra_metadata,
                    timestamps=timestamps,
                    timestamps_unit=timestamps_unit,
                    timestamp_extra_metadata=timestamp_extra_metadata,
                ),
            )
        except Exception as e:
            raise AssonantDataHandlerFactoryError("Failed to create Axis DataHandler with TimeSeries data.") from e

        return new_axis


class AssonantEntryFactory:
    """Assonant Entry Factory Class.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Entry objects.
    """

    @staticmethod
    def create_entry(entry_name: str, beamline_name: BeamlineName):
        """Create Assonant Entry from a given Beamline.

        Args:
            acquisition_moment (AcquisitionMoment): Enum that defines Acquisition Moment related to entry.
            beamline_name (BeamlineName): BeamlineName enum that defines which Beamline the entry is related to.

        Returns:
            Entry: Assonant Entry for the respective BeamlineName.
        """
        try:
            beamline = AssonantComponentFactory.create_component_by_class_type(Beamline, beamline_name.value)
            beamline.create_and_add_field(name="name", value=beamline_name.value)
            entry = Entry(name=entry_name, beamline=beamline)
        except Exception as e:
            raise AssonantEntryFactoryError(
                f"An Error ocurred when creating an Entry for Beamline '{beamline_name}'."
            ) from e
        return entry



"""Assonant Bluesky submodule exceptions.

This submodule defines Exceptions used over the Assonant Bluesky submodule.
"""

from .bluesky_data_parser_exception import BlueskyDataParserError
from .bluesky_data_parser_factory_exception import BlueskyDataParserFactoryError

__all__ = ["BlueskyDataParserError", "BlueskyDataParserFactoryError"]



from typing import Union

from assonant._kafka.producer import Producer
from assonant.data_classes import AssonantDataClass
from assonant.data_classes.events import AssonantEvent

from ._assonant_data_sender_interface import IAssonantDataSender
from .exceptions import AssonantDataSenderError


class KafkaDataSender(IAssonantDataSender):
    """Kafka Data Sender.

    Data sender specialized on sending AssonantDataClasses over a Kafka
    data broker
    """

    def __init__(self, configs: dict):
        self._parse_configs(configs)
        producer_configs = self._mount_producer_configs_dict()
        self.producer = Producer(configs=producer_configs)

    def send_data(self, payload: Union[AssonantDataClass, AssonantEvent]):
        """Method for sending data through a Kafka topic.

        Args:
            payload (Union[AssonantDataClass, AssonantEvent]): AssonantDataClass or AssonantEvent
            that will be send.
        """
        # Serialize payload as json and send it over topic
        self.producer.produce(topic=self.topic, value=payload.json())
        self._close()

    def _close(self):
        """Close Kafka Producer instance."""
        self.producer.close()

    def _parse_configs(self, configs: dict) -> None:
        """Parse configuration dictionary to extract configuration settings from it.

        Parsed data should be set to class properties, not returned by the method.

        Args:
            configs (dict): Dictionary containing configuration settings which will be parsed.
        """
        self._validade_configs(configs)

        self.server = configs["server"]
        self.topic = configs["topic"]

    def _validade_configs(self, configs: dict) -> None:
        """Validate if all mandatory configurations were passed in the configuration.

        Args:
            configs (dict): Dictionary containing configuration settings which will be parsed.

        Raises:
            AssonantDataSenderError: Mandatory field is missing in configuration data.
        """
        if "server" not in configs.keys():
            raise AssonantDataSenderError("Mandatory field 'server' missing on " "config data!")
        elif "topic" not in configs.keys():
            raise AssonantDataSenderError("Mandatory field 'topic' missing on " "config data!")

    def _mount_producer_configs_dict(self) -> dict:
        """Format configuration settings to specific format Kafka Producer expects.

        Returns:
            dict: Formatted configuration dictionary.
        """
        kafka_configs = {}
        kafka_configs["bootstrap.servers"] = self.server

        return kafka_configs



"""Assonant metadata retriever exceptions.

This submodule defines Exceptions used all over the metadata retriever module and its submodules.
"""

from .metadata_retrievers_exceptions import AssonantMetadataRetrieverError

__all__ = ["AssonantMetadataRetrieverError"]



"""Assonant Bluesky.

Assonant toolkit for dealing with tasks related to Bluesky.
"""

from .data_parser import DocumentsDataParser
from .factories import DataParserFactory

__all__ = ["DocumentsDataParser", "DataParserFactory"]



"""Assonant Entry data class."""

from typing import Dict, Generator, Iterable, List, Optional, Type, Union

import numpy as np

from .assonant_data_class import AssonantDataClass
from .components import Beamline
from .data_handlers import DataField, DataHandler, TimeSeries
from .exceptions.data_classes_exceptions import (
    DataHandlerInsertionError,
    DataHandlerTakeageError,
    FieldCreationError,
    FieldInsertionError,
)


class Entry(AssonantDataClass):
    """Data classes that wraps data into a logical/temporal scope related to the experiment.

    Entries are used to group and represent data in a defined temporal/logical scope of the
    experimen. Every Entry instance must also have an Beamline object to identify with at
    least the beamline name it is related to.
    """

    beamline: Beamline
    name: str
    fields: Optional[Dict[str, DataHandler]] = {}

    def add_field(self, name: str, new_field: DataField):
        """Add new data field to component.

        Args:
            name (str): Field name
            data_field (DataField): DataField DataHandler object.

        Raises:
            FieldInsertionError: Failure on inserting DataField to Component fields dict.
        """
        try:
            if name not in self.fields:
                self.fields[name] = new_field
            else:
                # Generate a new name with a index to avoid equal names
                new_name = self._generate_new_valid_name(names_list=self.fields, name=name)
                self.fields[new_name] = new_field
        except Exception:
            raise FieldInsertionError(f"Failed to add '{name}' field to {self.name} Component.")

    def create_and_add_field(
        self,
        name: str,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Create and add new data field.

        Args:
            name (str): Field name.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]], optional): Dict
            containing any aditional metadata related to collected data. Defaults to {}.

        Raises:
            FieldCreationError: Failure on creating new DataField.
            FieldInsertionError: Failure on inserting DataField into Component fields dict.
        """
        from .factories import AssonantDataHandlerFactory

        try:
            new_field = AssonantDataHandlerFactory.create_data_field(
                value=value,
                unit=unit,
                extra_metadata=extra_metadata,
            )
        except Exception as e:
            raise FieldCreationError(
                f"Failed to create '{name}' DataField Data Handler for {self.name} Component."
            ) from e
        try:
            self.add_field(name=name, new_field=new_field)
        except FieldInsertionError as e:
            raise e

    def add_timeseries_field(self, name: str, new_time_series_field: TimeSeries):
        """Add new data field to component that was collected as a TimeSeries.

        Args:
            name (str): Field name
            new_time_series_field (TimeSeries): TimeSeries DataHandler object.

        Raises:
            FieldInsertionError: Failure to insert TimeSeries to Component fields dict.
        """
        try:
            if name not in self.fields:
                self.fields[name] = new_time_series_field
            else:
                # Generate a new name with a index to avoid equal names
                new_name = self._generate_new_valid_name(names_list=self.fields, name=name)
                self.fields[new_name] = new_time_series_field
        except Exception:
            raise FieldInsertionError(f"Failed to add '{name}' TimeSeries field to {self.name} Component.")

    def create_and_add_timeseries_field(
        self,
        name: str,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        timestamps: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Create and add new data field to component that data was collected as a TimeSeries.

        Args:
            name (str): Field name.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            timestamps (Union[int, float, str, List, Type[np.ndarray], None]): Timestamps related to data collected
            from the field.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing any aditional metadata related to collected data. Defaults to {}.
            timestamps_unit (Optional[str], optional): Measurement unit related to
            timestamp field. Defaults to None.
            timestamp_extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing extra metadata about timestamps field. Defaults to {}.

            Raises:
                FieldCreationError: Failure on creating new DataField.
                FieldInsertionError: Failure on inserting DataField into Component fields dict.
        """
        from .factories import AssonantDataHandlerFactory

        try:
            new_field = AssonantDataHandlerFactory.create_timeseries_field(
                value=value,
                unit=unit,
                extra_metadata=extra_metadata,
                timestamps=timestamps,
                timestamps_unit=timestamps_unit,
                timestamp_extra_metadata=timestamp_extra_metadata,
            )
        except Exception as e:
            raise FieldCreationError(
                f"Failed to create '{name}' TimeSeries Data Handler for {self.name} Component."
            ) from e
        try:
            self.add_timeseries_field(name=name, new_time_series_field=new_field)
        except FieldInsertionError as e:
            raise e

    def add_data_handler(self, name: str, new_data_handler: DataHandler):
        """Add new DataHandler to its specific dictionary within Component class.

        This method is a generic way to add any type of DataHandler to a component. It just
        call the specific underlying method to add that type of DataHandler to the Component.

        Args:
            name (str): name (str): Name attributed to passed DataHandler.
            new_data_handler (DataHandler): Any valid DataHandler object.

        Raises:
            DataHandlerInsertionError: Raised when the DataHandler type is not supported or an error occurs
            inside the specific method to add the DataHandlers to the component based in its type.
        """
        try:
            if isinstance(new_data_handler, DataField):
                self.add_field(name=name, new_field=new_data_handler)
            elif isinstance(new_data_handler, TimeSeries):
                self.add_timeseries_field(name=name, new_time_series_field=new_data_handler)
            else:
                raise DataHandlerInsertionError(
                    f"Adding DataHandler of type: {type(new_data_handler)} is not supported yet."
                )
        except Exception as e:
            raise DataHandlerInsertionError(
                f"Failed to add DataHandler of type {type(new_data_handler)} to {self.name} Component."
            ) from e

    def take_data_handlers_from(self, entry: Union["Entry", List["Entry"]]):
        """Take DataHandlers from within an Entry or List of Entrys and add to itself.

        Args:
            entry (Union[Entry, List[Entry]]): Entry or List of Entrys which DataHandlers will be taken.

        Raises:
            DataHandlerTakeageError: Raised when an error occurs when taking a DataHandler and trying to add to itself.
        """

        if not isinstance(entry, List):
            entry_list = [entry]
        else:
            entry_list = entry

        for target_entry in entry_list:
            for field_name, field_data_handler in target_entry.fields:
                try:
                    self.add_data_handler(name=field_name, data_handler=field_data_handler)
                except Exception as e:
                    raise DataHandlerTakeageError(f"Failed when trying to take '{field_name}' field from Entry") from e

    def _generate_new_valid_name(self, names_list: Iterable, name: str) -> str:
        """Generate a new name that don't exist on passed names list.

        Args:
            names_list (_type_): Iterable containing already existing names.
            name (str): Current name that will be changed to fit passed list.

        Returns:
            str: New unique name based on the passed name that don't exist on passed names list.
        """
        # Generate a new name with a index to avoid equal names
        new_name_generator = self._indexed_name_generator(name)
        new_name = next(new_name_generator)
        while new_name in names_list:
            new_name = next(new_name_generator)
        return new_name

    def _indexed_name_generator(self, name: str) -> Generator[str, None, None]:
        """Generator for creating indexed names.

        Args:
            name (str): Reference name to be indexed by generator.

        Yields:
            Generator[str, None, None]: New indexed name based on passed name.
        """
        i = 1
        while True:
            new_name = f"{name} ({i})"
            yield new_name
            i += 1



"""Assonant ExternalLink class."""
from .data_handler import DataHandler


class ExternalLink(DataHandler):
    """Data class to handle a link that references a specific data from an external file."""

    name: str
    target_path: str
    filepath: str



from typing import List, Set, Tuple, Union

from assonant.metadata_retriever import AssonantMetadataRetriever

from .components import Beamline, Component
from .entry import Entry
from .exceptions import AssonantHierarchizerError
from .factories import AssonantComponentFactory


class Hierarchizer:
    """Assonant Data Class Hierarchizer.

    Object responsible to automate the creation and mounting of the correct Components
    hierarchy based in a given schema.
    """

    def __init__(self, schema_file_path: str, metadata_retriever: AssonantMetadataRetriever = None):
        """Assonant Data Class Hierarchizer Constructor.

        Args:
            schema_file_path (str): Path to schema file.
            metadata_retriever (AssonantDataRetriever, optional): An already instantiate AssonantMetadataRetriever
            to be used. This can be used to avoid unneeded creation of new AssonantDatRetriever,
            which may generate some overhead due to accessing and parsing data source file.
            Defaults to None.
        """
        self.schema_file_path = schema_file_path

        if metadata_retriever is not None:
            self.metadata_retriever = metadata_retriever
        else:
            try:
                self.metadata_retriever = AssonantMetadataRetriever(self.schema_file_path)
            except Exception as e:
                raise AssonantHierarchizerError(
                    f"Failed to instantiate an AssonantDataRetriever for the data class Hierarchizer with {self.schema_file_path}"
                ) from e

    def _mount_tree_node_set(self, component_info_tuple: Tuple[str, Union[str, None]]) -> Set[str]:
        """Mount a set of tree nodes that connects component_info_tuple to its root component_info_tuple.

        PS: component_info_tuples may be created to allow achieving the root node. Due to that
        a metadata_retriever is needed to allow searching for missing nodes in the schema.

        Args:
            component_info_tuple (Tuple[str, Union[str, None]]): Tuple containing Component info.

        Returns:
            set[str]: Set containing all component_info_tuples that connect the passed component_info_tuple
            to its root node component_info_tuple.
        """
        # Auxiliar index naming for better code comprehension
        # COMPONENT_NAME = 0
        # COMPONENT_CLASS = 1
        COMPONENT_PARENT = 2

        # Check if component is the root node (No parent component)
        if component_info_tuple[COMPONENT_PARENT] is not None:
            # TODO: POSSIBLE ENHANCEMENT: CHECK IF COMPONENT INFO ALREADY DONT EXIST ON COMPONENT_INFO_SET
            parent_component_info = self.metadata_retriever.get_component_info(component_info_tuple[COMPONENT_PARENT])
            parent_component_info_tuple = (
                parent_component_info["component_info"]["name"],
                parent_component_info["component_info"]["class"],
                (
                    parent_component_info["component_info"]["subcomponent_of"]
                    if "subcomponent_of" in parent_component_info["component_info"].keys()
                    else None
                ),
            )
            set_of_tree_nodes = self._mount_tree_node_set(parent_component_info_tuple)
            set_of_tree_nodes.add(component_info_tuple)

        else:
            set_of_tree_nodes = set()
            set_of_tree_nodes.add(component_info_tuple)

        return set_of_tree_nodes

    def hierarchize_components(self, components: List[Component]) -> List[Component]:
        """Hierarchize passed Components according to subcomponents metadata.

        Args:
            components (List[Component]): List of Components that will be hierarchized.

        Returns:
            List[Component]: List of hierarchized Components.
        """

        # =============== UNION-FIND based algoritm solution ===============

        # Problem definition:
        # Data may be acquired in a sparse way where not all Components/SubComponents have data
        # collected and are required to be collected in all experiments.
        # To avoid situations where a Subcomponent have data collect but its parent Component not,
        # and the Subcomponent would be store in a different hierarchy
        # level when compared to situation where its parent Component has information collected,
        # this algorithm was developed. The idea is, without creating all
        # Components and Subcomponents possible and letting many empty if they don't have
        # data collected, guarantee that the collected data will be fit
        # the same hierarchy level no matter the situation. THis allow a standardization in the data
        # schema, and still allow non-collected data to simply not appear.
        #
        # Overall Idea of the algorithm
        # To achieve that, this algorithm get all collected Components, create sets with their
        # information an start to merge them until achieve disjoint sets, meaning
        # a root Component was found. The data retriever is needed here to query the reference schema
        # to allow filling empty roles in the hierarchy in order to allow
        # every Subcomponent to be connected to its root Component.
        # Below there are each step of the algorithm and a brief explanation
        # ===================================================================

        components_references = {}
        components_info_set = set()
        unchanged_components = []
        hierarchized_components = []

        # Step 0: Prepare data structure for Union-Find based algorithm
        for component in components:
            # Ignore Entry and Beamline!! They are special cases in the hierarchy!!!
            if isinstance(component, Entry) or isinstance(component, Beamline):
                unchanged_components.append(component)
            else:
                # Store reference to component in dict
                components_references[component.name] = component

                # Retrieve information about the component to get 'subcomponent of' information
                component_retrieved_info = self.metadata_retriever.get_component_info(component.name)

                # Add component information to set that will be used on further steps
                component_name = component_retrieved_info["component_info"]["name"]
                component_class = component_retrieved_info["component_info"]["class"]
                component_parent = (
                    component_retrieved_info["component_info"]["subcomponent_of"]
                    if "subcomponent_of" in component_retrieved_info["component_info"].keys()
                    else None
                )
                components_info_set.add((component.name, component_class, component_parent))

        forest_of_tree_node_sets = []

        # Step 1: Mount Each collected Component minimum hierarchy tree that connects each Component
        # to its respective root Component
        for component_info_tuple in components_info_set:
            forest_of_tree_node_sets.append(self._mount_tree_node_set(component_info_tuple))

        # print("Step1:", forest_of_tree_node_sets)

        # STEP 2: Unite tree sets until they are all disjoints
        solved_tree_node_sets = []

        while forest_of_tree_node_sets != []:

            # Get one node set
            chosen_tree_node_set = forest_of_tree_node_sets.pop()
            intersecting_set_found = False

            # Check if the chosen set intersects other existing set
            for tree_node_set in forest_of_tree_node_sets:
                if not chosen_tree_node_set.isdisjoint(tree_node_set):
                    # A intersecting set was found!

                    # Unite both sets
                    chosen_tree_node_set = chosen_tree_node_set.union(tree_node_set)

                    # Remove from list set that was united with chosen set to avoid duplication
                    forest_of_tree_node_sets.remove(tree_node_set)

                    # Add the United set to the pool of sets
                    forest_of_tree_node_sets.append(chosen_tree_node_set)
                    intersecting_set_found = True
                    break

            # If no intersecting set was found, it means that chosen set is solved
            if not intersecting_set_found:
                solved_tree_node_sets.append(chosen_tree_node_set)

        # print("Step2:", solved_tree_node_sets)

        # Step 3: Instantiate Components or reuse already instantiated ones and put subcomponents inside their parents
        for solved_tree_node_set in solved_tree_node_sets:
            root_node = None

            # Create all Components
            for node_tuple in solved_tree_node_set:
                component_name, component_class_name, component_parent_name = node_tuple
                # Check if Component was a intermediate Component in the hierarchy which no information was collected
                if component_name not in components_references.keys():
                    # Instantiate Component to fill hierarchy gaps
                    components_references[component_name] = AssonantComponentFactory.create_component_by_class_name(
                        component_class_name, component_name
                    )

            # Create Components/Subcomponents Hierarchy
            for node_tuple in solved_tree_node_set:
                component_name, component_class_name, component_parent_name = node_tuple
                if component_parent_name is None:
                    # Save reference for root node
                    root_node = components_references[component_name]
                else:
                    # Add component as subcomponent of its parent
                    components_references[component_parent_name].add_subcomponent(components_references[component_name])

            hierarchized_components.append(root_node)

        # print("Step3:", hierarchized_components)

        # Step 4: Concatenate Unchanged Components with Hierarchized Components

        # print("step4:", unchanged_components + hierarchized_components)

        return unchanged_components + hierarchized_components



"""Assonant Bending Magnet component data class."""

from .component import Component


class BendingMagnet(Component):
    """Data class to handle all data required to define a bending magnet insertion device."""



"""Assonant Granite base component data class."""

from .component import Component


class GraniteBase(Component):
    """Data class to handle all data required to define granite base where many components are mounted over it."""



"""Assonant Data Classes.

Assonant data classes defines standard grouping classes for acquiring sending,
manipulating and storing data over Assonant modules.

All classes inherits from AssonantDataClass and can be split mainly in 3 types:

1. Components: Data classes responsible to grouping information in logical scopes (e.g: Data from
a detector is stored inside a Detector class, while sample information in a class called Sample)
2. Data Handler: Data classes responsible for standardizing how data is handled inside other
Assonant data classes.
3. Entry: Special class used to group all information related to a specific stage of an
Experiment.

To import and uses data classes from each type, refer to its specific submodule as shown below:

from .assonant_data_class.<sub_module_name> import <data_class_name>, ...
"""

from .assonant_data_class import AssonantDataClass
from .entry import Entry
from .factories import (
    AssonantComponentFactory,
    AssonantDataHandlerFactory,
    AssonantEntryFactory,
)
from .hierarchizer import Hierarchizer

__all__ = [
    "AssonantComponentFactory",
    "AssonantDataClass",
    "AssonantDataHandlerFactory",
    "AssonantEntryFactory",
    "Entry",
    "Hierarchizer",
]



import json
from typing import Any, Dict, List, Tuple, Union

from event_model import DocumentRouter

from .enums import EventStreamName
from .exceptions import BlueskyDataParserError


class DocumentsDataParser(DocumentRouter):
    def __init__(self, bluesky_documents_list: List[List[Union[str, Dict[str, Any]]]], *args, **kwargs):
        """Constructor to instantiate a Bluesky Documents DataParser.

        Args:
            bluesky_documents_list (List[List[Union[str, Dict[str, Any]]]]): List of bluesky documents data
            presented also as a list. The list with the document data has 2 positions, the 1st is the bluesky
            identification for the document (start, event, stop, descriptor, ...) and the 2nd position the
            dict containing the document data.
        Raises:
            BlueskyDataParserError: Failure to read bluesky documents.
        """
        super().__init__(*args, **kwargs)

        self._start_document = None
        self._descriptor_documents = {}

        # Stream name --> Subdevice configurations
        self._configurations = {}
        # Stream name --> Device data keys
        self._data_keys = {}
        # Stream name --> Device data key --> (Device data reading, Reading timestamp)
        self._event_data = {}

        try:
            print("Init DocumentsDataParser")
            # Iterate over json data correctly parse data based on event type
            for event_name, document_data_dict in bluesky_documents_list:
                print(event_name, document_data_dict)
                self(event_name, document_data_dict)
        except Exception as e:
            raise BlueskyDataParserError(
                f"Failed to read bluesky documents from structure: {bluesky_documents_list}"
            ) from e

    def start(self, doc: Dict):
        """DocumentRouter method related to processing bluesky START documments

        Args:
            doc (Dict): Dictionary containing bluesky documment data.
        """
        print("========== START ==========")
        print(doc)
        print("===========================\n")
        self._start_document = doc

    def descriptor(self, doc: Dict):
        """DocumentRouter method related to processing bluesky DESCRIPTOR documments

        Args:
            doc (Dict): Dictionary containing bluesky documment data.
        """
        print("========== DESC ==========")
        print(doc)
        print("===========================\n")
        self._descriptor_documents[doc["uid"]] = doc

        self._configurations[doc["name"]] = doc["configuration"]
        self._data_keys[doc["name"]] = doc["data_keys"]
        # print("DESC DOCS")
        # print(self._descriptor_documents)

    def event(self, doc: Dict):
        """DocumentRouter method related to processing bluesky EVENT documments

        Args:
            doc (Dict): Dictionary containing bluesky documment data.
        """
        print("========== EVENT ==========")
        print(doc)
        print("===========================\n")
        stream_name = self._descriptor_documents[doc["descriptor"]]["name"]

        if stream_name not in self._event_data:
            self._event_data[stream_name] = {}

        for dev_name, val in doc["data"].items():
            if dev_name not in self._event_data[stream_name]:
                self._event_data[stream_name][dev_name] = []

            self._event_data[stream_name][dev_name].append((val, doc["timestamps"][dev_name]))

    def all_data_keys(self) -> Dict[str, List]:
        """Return all dictionary keys to access data fields from each stream into the data dictionary.

        Returns:
            Dict[str, List]: Dictionary containing the stream name and a List with its respetive data_keys
        """
        data_keys = {}
        for stream, devices in self._data_keys.items():
            for device_data_key, device_values in devices.items():
                if stream not in data_keys:
                    data_keys[stream] = []
                data_keys[stream].append(device_data_key)
        return data_keys

    def data_key_to_pv_name(self, data_key: str) -> str:
        """Convert data_key to the respective PV name related to it.

        Args:
            data_key (str): data key from data dictionary that will be converted

        Returns:
            str: PV Name respective to passed data_key
        """
        pv_names = {}
        for stream, devices in self._data_keys.items():
            for device_data_key, device_values in devices.items():
                if device_data_key == data_key:
                    pv_names[stream] = device_values["source"][3:]
                    break
        return pv_names

    def pv_name_to_data_key(self, pv_name: str) -> str:
        """Convert PV name to the respective data_key related to it.

        Args:
            pv_name (str): PV Name that will be converted

        Returns:
            str: Data key respective to passed PV Name
        """
        data_keys = {}
        for stream, devices in self._data_keys.items():
            for device_data_key, device_values in devices.items():
                if pv_name in device_values["source"]:
                    data_keys[stream] = device_data_key
                    break

        return data_keys

    def query_data_by_stream_name(
        self, event_stream_name: EventStreamName, with_config: bool
    ) -> Dict[str, List[Tuple[Any, float]]]:
        """Return all information from bluesky docs from a specific event dtream organized as PV name -> Value.

        The organization structure returned is the following:
        {
            PV_NAME_1: [(PV_VALUE_1, PV_VALUE_1_TIMESTAMP), (PV_VALUE_2, PV_VALUE_2_TIMESTAMP) ...],
            PV_NAME_2: [(PV_VALUE_1, PV_VALUE_1_TIMESTAMP)],
            ...
        }

        PS: PV values are ALWAYS inside a List, even if it is a single value. It is also ALWAYS
        returned in a Tuple where the first position is the acquired value and the second the
        acquisition timestamp.

        Args:
            event_stream_name (EventStreamName): Query's target event stream name.
            with_config (bool): Flag to control if data will add configuration
            information or not.

        Returns:
            Dict[str, Dict[str, List]]: Dictionary with data following structure described above. If
            no data is found it returns an empty dict.
        """
        # Stream name --> PV prefix --> Collected data
        data = {}

        data_keys = self.all_data_keys()
        if len(data_keys) == 0 or event_stream_name.value not in data_keys.keys():
            return data

        data_keys = data_keys[event_stream_name.value]

        for data_key in data_keys:
            pv_name = self.data_key_to_pv_name(data_key)[event_stream_name.value]

            # print()
            # print(f"  Data key: {data_key}")
            # print(f"  PV: {pv_name}")

            # TODO: DEVELOP A WAY TO RETURN THE CONFIGURATIONS AMONG THE COLLECTED DATA
            # if with_config:
            #     device_config = self._data_keys[stream][data_key]

            #     print("  Configuration:")
            #     print("\n".join(f"    {k}: {v}" for k, v in device_config.items()))

            #     if data_key in self._configurations[stream]:
            #         for subsignal_name, subsignal_data in self._configurations[stream][data_key][
            #             "data_keys"
            #         ].items():
            #             if subsignal_name == data_key:
            #                 continue

            #             print()
            #             print(f"    Sub-signal '{subsignal_name}':")
            #             print("\n".join(f"      {k}: {v}" for k, v in subsignal_data.items()))

            # print("  Data:")
            device_data = self._event_data[event_stream_name.value]
            # for value, timestamp in device_data[data_key]:
            # print(f"      {value} ({timestamp})")

            data[pv_name] = device_data[data_key]
        return data

    def query_all_data(self, with_config: bool) -> Dict[str, Dict[str, List[Tuple[Any, float]]]]:
        """Return all information from bluesky docs organized as Stream -> PV name -> Value.

        The organization structure returned is the following:
        {
            STREAM_NAME_1: <Structure from query_data_by_stream_name() method>,
            STREAM_NAME_2: <Structure from query_data_by_stream_name() method>,
            ...
        }

        PS: PV values are ALWAYS inside a List, even if it is a single value. It is also ALWAYS
        returned in a Tuple where the first position is the acquired value and the second the
        acquisition timestamp.

        Args:
            with_config (bool): Flag to control if data will add configuration
            information or not.

        Returns:
            Dict[str, Dict[str, List]]: Dictionary with data following structure described above. If
            no data is found it returns an empty dict.
        """
        # Stream name --> PV prefix --> Collected data
        data = {}

        data_keys = self.all_data_keys()
        if len(data_keys) == 0:
            return data

        for stream in data_keys.keys():
            # print(f"Stream '{stream}':")
            data[stream] = {}
            event_stream_name = EventStreamName._value2member_map_[stream]
            stream_data = self.query_data_by_stream_name(event_stream_name, with_config=with_config)
            data[stream] = stream_data
        return data

    def load_json(self, path: str) -> Dict[str, Any]:
        """Load data from passed json file

        Args:
            path (str): File path to json file containing data collect from bluesky.

        Returns:
            Dict[str, Any]: Dictionary containing retrieved data from json file.
        """
        contents = None
        with open(path, "r") as _f:
            contents = json.load(_f)

        return contents



"""Assonant data parsers for data loggers.

This submodule defines Data Parsers objects to be used on Data Logger in order to interpretate
different types of data mapping/configuration files.
"""

from .assonant_data_retriever import AssonantDataRetriever

__all__ = [
    "AssonantDataRetriever",
]



"""Tests focused on validating Assonant Components/Entry methods that add information to it"""

import pytest

from assonant.data_classes.data_handlers import Axis, DataField
from assonant.data_classes.enums import TransformationType
from assonant.data_classes.factories import AssonantDataHandlerFactory

from ..utils import (
    assert_axis_content,
    assert_axis_equality,
    assert_component_equality,
    assert_data_field_content,
    assert_data_fields_equality,
    get_fixture_values,
)
from .test_data_handler_factory import create_test_params_combinations


@pytest.mark.parametrize("transformation_type", [t_type for t_type in TransformationType])
@pytest.mark.parametrize("value, unit, metadata", create_test_params_combinations(Axis))
def test_add_position(value, unit, metadata, transformation_type, example_base_component, request):
    """
    Validate that add_position method from Components correctly
    add Axis object to its internal positions dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must add Axis within its position dictionary
        - Added Axis must be identified with the same name it was given during add_position method
        - Axis object within Component positions dictionary must be equivalent to how it was before insertion
    """

    # To avoid creating lots of different fixtures, manually create Axis DataHandler using Factory
    # if this test passes, other tests related to adding Axis, don't need to deal with all data types.
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)

    axis = AssonantDataHandlerFactory.create_axis(
        transformation_type=transformation_type,
        value=input_value,
        unit=input_unit,
        extra_metadata=input_metadata,
    )

    axis_name = "x"
    example_base_component.add_position(axis_name, axis.model_copy())

    # Assert Axis was add to positions dict
    assert len(example_base_component.positions.items()) == 1

    # Assert correct name was given to Axis within positions dict and that it remained unchanged.
    assert_axis_equality(example_base_component.positions[axis_name], axis)


def test_add_position_with_name_conflict(example_base_component, example_translation_axis, example_rotation_axis):
    """
    Validate that add_position method from Components correctly add Axis object to its internal
    positions dictionary and handles name conflicts.

    Expected Behavior:
        - Both added Axis must be within Component positions internal dict
        - The first added Axis must not suffer any change from how it was before
        - Any subsequent Axis added that has the same name with another already existing Axis from the poisitions
        dict from Component must have its name automatically updated to a new unique name by adding an index in
        front the original passed name (e.g: <axis_name> (1))
    """
    axis_name = "x"
    example_base_component.add_position(axis_name, example_translation_axis.model_copy())
    example_base_component.add_position(axis_name, example_rotation_axis.model_copy())

    # Assert both Axis objects were add to positions dict
    assert len(example_base_component.positions.items()) == 2

    # 1st Axis name and content must be unchanged
    assert_axis_equality(example_base_component.positions[axis_name], example_translation_axis)

    # 2nd Axis name must be the same as the 1st with an index due to name conflict and its content must be unchanged
    assert_axis_equality(example_base_component.positions[axis_name + " (1)"], example_rotation_axis)


@pytest.mark.parametrize("transformation_type", [t_type for t_type in TransformationType])
@pytest.mark.parametrize("value, unit, metadata", create_test_params_combinations(DataField))
def test_create_and_add_position(value, unit, metadata, transformation_type, example_base_component, request):
    """
    Validate that create_and_add_position method from Components correctly
    create and add Axis object to its internal positions dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must create a new Axis with passed data
        - Base Component must add the new created Axis within its positions dictionary
        - Added Axis must be identified with the same name it was given during create_and_add_position method
        - Axis object created by the Component must preserve passed data values.
    """
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)
    input_transformation_type = transformation_type

    axis_name = "x"
    example_base_component.create_and_add_position(
        axis_name, input_transformation_type, input_value, input_unit, input_metadata
    )

    # Assert created Axis was add to positions dict
    assert len(example_base_component.positions.items()) == 1

    # Assert correct name was given to Axis within positions dict and that it remained unchanged
    assert_axis_content(
        example_base_component.positions[axis_name], input_value, input_unit, input_metadata, input_transformation_type
    )


def test_create_and_add_position_with_name_conflict(
    example_float_1d_array, example_int_1d_array, example_unit, example_metadata, example_base_component
):
    """
    Validate that create_and_add_position method from Components correctly
    create and add Axis object to its internal positions dictionary and handles name conflict.

    Expected Behavior:
        - Both Axis DataHandlers must be created
        - Both created Axis must be add to Component positions dictionary
        - The first create and added Axis must have its name and content preserved
        - The second create and added Axis must have its content preserved but name updated
        to a new unique name by adding an index in front of the original anem (e.g: <axis_name> (1))
    """
    axis_name = "x"
    example_base_component.create_and_add_position(
        axis_name, TransformationType.TRANSLATION, example_float_1d_array, example_unit, example_metadata
    )
    example_base_component.create_and_add_position(
        axis_name, TransformationType.ROTATION, example_int_1d_array, example_unit, example_metadata
    )

    # Assert both created Axis were add to positions dict
    assert len(example_base_component.positions.items()) == 2

    # 1st Axis name and content must be unchanged
    assert_axis_content(
        example_base_component.positions[axis_name],
        example_float_1d_array,
        example_unit,
        example_metadata,
        TransformationType.TRANSLATION,
    )

    # 2nd Axis name must be the same as the 1st with an index due to name conflict and its content must be unchanged
    assert_axis_content(
        example_base_component.positions[axis_name + " (1)"],
        example_int_1d_array,
        example_unit,
        example_metadata,
        TransformationType.ROTATION,
    )


def test_add_subcomponent(example_base_component, example_subcomponent):
    """
    Validate that add_subcomponent method from Components correctly
    add subcomponents to its internal dictionary.

    Expected Behavior:
        - Both added subcomponents must be within base_component internal sub_components dict
        - Added sub_components must be exactly equal to how it was before being add
    """
    example_subcomponent_2 = example_subcomponent.model_copy()
    example_subcomponent_2.name = "new unique name"

    # Add copies to avoid that internal manipulation changes external data and causing possible False-Positive result.
    example_base_component.add_subcomponent(example_subcomponent.model_copy())
    example_base_component.add_subcomponent(example_subcomponent_2.model_copy())

    assert len(example_base_component.subcomponents.items()) == 2
    assert_component_equality(example_base_component.subcomponents[example_subcomponent.name], example_subcomponent)
    assert_component_equality(example_base_component.subcomponents[example_subcomponent_2.name], example_subcomponent_2)


def test_add_subcomponent_with_name_conflict(example_base_component, example_subcomponent):
    """
    Validate that add_subcomponent method from Components correctly
    handle cases where subcomponent with conflicting name are add.

    Expected Behavior:
        - Both added subcomponents must be within base_component internal sub_components dict
        - The first added sub_components must not suffer any change from how it was before
        - Any subsequent sub_component added that has the same name with another already existing subcomponent
        must have its name automatically updated to a new unique name by adding an index in front the original
        name (e.g: sub_component (1))
    """
    example_subcomponent_2 = example_subcomponent.model_copy()

    example_base_component.add_subcomponent(example_subcomponent.model_copy())
    example_base_component.add_subcomponent(example_subcomponent_2.model_copy())

    assert len(example_base_component.subcomponents.items()) == 2

    # Assert that 1st subcomponent was totally unchanged
    assert_component_equality(example_base_component.subcomponents[example_subcomponent.name], example_subcomponent)

    with pytest.raises(AssertionError):
        # It is expected the assertion to fail, as subcomponent name should have been renamed,
        # otherwise name conflict handling failed
        assert_component_equality(
            example_base_component.subcomponents[example_subcomponent_2.name + " (1)"], example_subcomponent_2
        )

    # Assert it was only Component names that were different and caused previous AssertionError
    renamed_example_subcomponent_2 = example_subcomponent_2.model_copy()
    renamed_example_subcomponent_2.name = example_base_component.subcomponents[
        example_subcomponent_2.name + " (1)"
    ].name
    assert_component_equality(
        example_base_component.subcomponents[example_subcomponent_2.name + " (1)"], renamed_example_subcomponent_2
    )


def test_add_list_of_subcomponents(example_base_component, example_subcomponent):
    """
    Validate that add_subcomponent method from Components correctly
    add list of subcomponents to its internal dictionary.

    Expected Behavior:
        - All added subcomponents must be within base_component internal sub_components dict
        - Added sub_components must be exactly equal to how it was before being add
    """
    example_subcomponent_2 = example_subcomponent.model_copy()
    example_subcomponent_3 = example_subcomponent.model_copy()
    example_subcomponent_2.name = "2nd subcomponent"
    example_subcomponent_3.name = "3rd subcomponent"

    list_of_subcomponents = [
        example_subcomponent.model_copy(),
        example_subcomponent_2.model_copy(),
        example_subcomponent_3.model_copy(),
    ]

    # Add copies to avoid that internal manipulation changes external data and causing possible False-Positive result.
    example_base_component.add_subcomponent(list_of_subcomponents)

    assert len(example_base_component.subcomponents.items()) == len(list_of_subcomponents)
    assert_component_equality(example_base_component.subcomponents[example_subcomponent.name], example_subcomponent)
    assert_component_equality(example_base_component.subcomponents[example_subcomponent_2.name], example_subcomponent_2)
    assert_component_equality(example_base_component.subcomponents[example_subcomponent_3.name], example_subcomponent_3)


def test_add_list_of_subcomponents_with_name_conflict(example_base_component, example_subcomponent):
    """
    Validate that add_subcomponent method from Components correctly handle cases where subcomponent
    with conflicting name exists inside subcomponents list being.

    Expected Behavior:
        - All added subcomponents must be within base_component internal sub_components dict
        - The first added sub_components must not suffer any change from how it was before
        - Any subsequent sub_component added that has the same name with another already existing subcomponent
        must have its name automatically updated to a new unique name by adding an index in front the original
        name (e.g: sub_component (1))
    """
    example_subcomponent_2 = example_subcomponent.model_copy()
    example_subcomponent_3 = example_subcomponent.model_copy()

    list_of_subcomponents = [
        example_subcomponent.model_copy(),
        example_subcomponent_2.model_copy(),
        example_subcomponent_3.model_copy(),
    ]

    # Add copies to avoid that internal manipulation changes external data and causing possible False-Positive result.
    example_base_component.add_subcomponent(list_of_subcomponents)

    assert len(example_base_component.subcomponents.items()) == len(list_of_subcomponents)

    # Assert that 1st subcomponent was totally unchanged
    assert_component_equality(example_base_component.subcomponents[example_subcomponent.name], example_subcomponent)

    with pytest.raises(AssertionError):
        # It is expected the assertion to fail, as subcomponent name should have been renamed,
        # otherwise name conflict handling failed
        assert_component_equality(
            example_base_component.subcomponents[example_subcomponent_2.name + " (1)"], example_subcomponent_2
        )

    # Assert it was only Component names that were different and caused previous AssertionError
    renamed_example_subcomponent_2 = example_subcomponent_2.model_copy()
    renamed_example_subcomponent_2.name = example_base_component.subcomponents[
        example_subcomponent_2.name + " (1)"
    ].name
    assert_component_equality(
        example_base_component.subcomponents[example_subcomponent_2.name + " (1)"], renamed_example_subcomponent_2
    )

    with pytest.raises(AssertionError):
        # It is expected the assertion to fail, as subcomponent name should have been renamed,
        # otherwise name conflict handling failed
        assert_component_equality(
            example_base_component.subcomponents[example_subcomponent_3.name + " (2)"], example_subcomponent_3
        )

    # Assert it was only Component names that were different and caused previous AssertionError
    renamed_example_subcomponent_3 = example_subcomponent_3.model_copy()
    renamed_example_subcomponent_3.name = example_base_component.subcomponents[
        example_subcomponent_3.name + " (2)"
    ].name
    assert_component_equality(
        example_base_component.subcomponents[example_subcomponent_3.name + " (2)"], renamed_example_subcomponent_3
    )


@pytest.mark.parametrize("value, unit, metadata", create_test_params_combinations(DataField))
def test_add_field(value, unit, metadata, example_base_component, request):
    """
    Validate that add_field method from Components correctly
    add DataField object to its internal fields dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must add DataField within its fields dictionary
        - Added DataField must be identified with the same name it was given during add_field method
        - DataField object within Component fields dictionary must be equivalent to how it was before insertion
    """

    # To avoid creating lots of different fixtures, manually create DataField DataHandler using Factory
    # if this test passes, other tests related to adding DataField, don't need to deal with all data types.
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)

    data_field = AssonantDataHandlerFactory.create_data_field(
        value=input_value,
        unit=input_unit,
        extra_metadata=input_metadata,
    )

    data_field_name = "data_field"
    example_base_component.add_field(data_field_name, data_field.model_copy())

    # Assert DataField was add to fields dict
    assert len(example_base_component.fields.items()) == 1

    # Assert correct name was given to DataField within fields dict and that it remained unchanged.
    assert_data_fields_equality(example_base_component.fields[data_field_name], data_field)


def test_add_field_with_name_conflict(
    example_base_component, example_int_1d_array_data_field, example_float_2d_array_data_field
):
    """
    Validate that add_field method from Components correctly add DataField object to its internal
    fields dictionary and handles name conflicts.

    Expected Behavior:
        - Both added DataFields must be within Component fields internal dict
        - The first added DataField must not suffer any change from how it was before
        - Any subsequent DataField added that has the same name with another already existing from the fields
        dict from Component must have its name automatically updated to a new unique name by adding an index in
        front the original passed name (e.g: <field_name> (1))
    """
    field_name = "field_name"
    example_base_component.add_field(field_name, example_int_1d_array_data_field.model_copy())
    example_base_component.add_field(field_name, example_float_2d_array_data_field.model_copy())

    # Assert both DataField objects were add to positions dict
    assert len(example_base_component.fields.items()) == 2

    # 1st DataField name and content must be unchanged
    assert_data_fields_equality(example_base_component.fields[field_name], example_int_1d_array_data_field)

    # 2nd DataField name must be the same as the 1st with a index due to name conflict and its content must be unchanged
    assert_data_fields_equality(example_base_component.fields[field_name + " (1)"], example_float_2d_array_data_field)


@pytest.mark.parametrize("value, unit, metadata", create_test_params_combinations(DataField))
def test_create_and_add_field(value, unit, metadata, example_base_component, request):
    """
    Validate that create_and_add_field method from Components correctly
    create and add DataField object to its internal fields dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must create a new DataField with passed data
        - Base Component must add the new created DataField within its fields dictionary
        - Added DataField must be identified with the same name it was given during create_and_add_field method
        - DataField object created by the Component must preserve passed data values.
    """
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)

    data_field_name = "data_field"
    example_base_component.create_and_add_field(data_field_name, input_value, input_unit, input_metadata)

    # Assert created DataField was add to fields dict
    assert len(example_base_component.fields.items()) == 1

    # Assert correct name was given to DataField within fields dict and that it remained unchanged
    assert_data_field_content(example_base_component.fields[data_field_name], input_value, input_unit, input_metadata)


def test_create_and_add_field_with_name_conflict(
    example_float_1d_array, example_int_1d_array, example_unit, example_metadata, example_base_component
):
    """
    Validate that create_and_add_field method from Components correctly
    create and add DataField object to its internal fields dictionary and handles name conflict.

    Expected Behavior:
        - Both DataFields must be created
        - Both created DataFields must be add to Component fields dictionary
        - The first create and added DataField must have its name and content preserved
        - The second create and added DataField must have its content preserved but name updated
        to a new unique name by adding an index in front of the original anem (e.g: <data_field_name> (1))
    """
    data_field_name = "data_field"
    example_base_component.create_and_add_field(data_field_name, example_float_1d_array, example_unit, example_metadata)
    example_base_component.create_and_add_field(data_field_name, example_int_1d_array, example_unit, example_metadata)

    # Assert both created DataFields were add to fields dict
    assert len(example_base_component.fields.items()) == 2

    # 1st DataField name and content must be unchanged
    assert_data_field_content(
        example_base_component.fields[data_field_name], example_float_1d_array, example_unit, example_metadata
    )

    # 2nd DataField name must be the same as the 1st with an index due to name conflict and
    # its content must be unchanged
    assert_data_field_content(
        example_base_component.fields[data_field_name + " (1)"], example_int_1d_array, example_unit, example_metadata
    )


# To test Axis failure and Data Field failure during insertion, pass wrong class to name param
# Test add subcomponent failure - exception raising
# Test add position failure - exception raising
# Test add field failure - exception raising
# Test create_and_add_position failure - exception raising
# Test create_and_add_field failure - exception raising


# Test create_and_add_timeseries_position failure - exception raising
# Test create_and_add_timeseries_position all valid data type combinations accepted
# Test create_and_add_timeseries_position duplicated
# Test create_and_add_field duplicated
# Test create_and_add_timeseries_field failure - exception raising
# Test create_and_add_timeseries_field for all valid data type combinations accepted
# Test create_and_add_timeseries_field duplicated
# Test create_data_handler generic method for all valid types accepted
# Test create_data_handler generic method duplicated for all types of data_handlers
# Test create_data_handler generic failed - exception raising
# Test take_data_handlers_from method
# Test generate_new_valid_name method
# Test indexed_name_generator method
# Test Components Constructors for all type of Components



"""General use utilitary methods"""

from typing import Any, Iterable, List, Union

import numpy as np

from assonant.data_classes.components import Component
from assonant.data_classes.data_handlers import Axis, DataField, TimeSeries
from assonant.data_classes.enums import TransformationType


def array_equal_safe_against_none_and_nan(
    a1: Union[List, np.typing.NDArray], a2: Union[List, np.typing.NDArray]
) -> bool:
    """Compare two arrays/lists if they are element-wise equal in a safe way against None and np.nan values.

    This method was create due to np.array_equals breaking if None value is passed when its equal_nan
    flag is set to True as np.isnan() methos used internally doesn't support None as input.

    Args:
        a1 (Union[List, np.typing.NDArray]): Array or List to be compared against a2.
        a2 (Union[List, np.typing.NDArray]): Array or List to be compared against a1.

    Returns:
        bool: True if a1 equals to a2 and False if not.
    """
    if a1 is Iterable and a2 is Iterable:
        if len(a1) != len(a2):
            # Different lengths, they are directly different.
            return False

    if np.array_equal(a1, a2):
        # Just a normal case of two equal arrays/lists.
        return True
    else:
        # Failure may be due to np.nan != np.nan so a specific comparsion need to be done
        # However, numpy method that does that breaks if None pass is passed. Due to that, an
        # element-wise comparison must be done and checking directly how to deal with specific cases.
        for v1, v2 in zip(a1, a2):
            if type(v1) is not type(v2):
                # Cover comparison between None and np.nan. ALso guarantee this will not happen
                # on underlying comparisons
                return False
            elif v1 is not None and v2 is not None:
                # Cover comparison using np.nan.
                if not np.array_equal(v1, v2, equal_nan=True):
                    return False
            else:
                # Cover comparison using None.
                if not np.array_equal(v1, v2):
                    return False

        # All elements were verified and no difference found, that means a1 == a2.
        return True


def compare_dicts(d1, d2):
    """Recursively compare two dictionaries, handling numpy arrays."""

    if d1 is not None and d2 is not None:
        if d1.keys() != d2.keys():
            # If dict keys are different, they are clearly different
            return False

        for key in d1:
            # Check equality of each dict element
            v1 = d1[key]
            v2 = d2[key]

            if isinstance(v1, dict) and isinstance(v2, dict):
                if not compare_dicts(v1, v2):
                    return False
            elif not array_equal_safe_against_none_and_nan(v1, v2):
                return False

    return True


def get_fixture_values(fixture: Union[str, None], request) -> Union[Any, None]:
    """Wrapper class to handle getting fixture values based in their names.

    Args:
        fixture (Union[str, None]): Fixture name or None value.
        request (pytest.FixtureRequest): request fixture object from pytest.

    Returns:
       Union[Any, None]: Retrieved fixture value if not None else returns None.
    """

    fixture_value = request.getfixturevalue(fixture) if fixture is not None else None

    return fixture_value


def assert_data_field_content(data_field: DataField, input_value: Any, input_unit: Any, input_metadata: Any):
    """Assert that DataField was correctly created based on input values.

    Args:
        data_field (DataField): DataField which assertion will be applied over.
        input_value (Any): Input value parameter during DataField creation.
        input_unit (Any): Input unit parameter during DataField creation.
        input_metadata (Any): Input metadata parameter during DataField creation.

    Raise:
        AssertionError: Raises if any assertion fails.
    """
    assert isinstance(data_field, DataField)
    assert array_equal_safe_against_none_and_nan(data_field.value, input_value)
    assert data_field.unit == input_unit
    assert compare_dicts(data_field.extra_metadata, input_metadata)


def assert_timeseries_content(
    timeseries: TimeSeries,
    input_value: Any,
    input_unit: Any,
    input_metadata: Any,
    input_timestamps: Any,
    input_timestamps_unit: Any,
    input_timestamps_metadata: Any,
):
    """Assert that TimeSeries was correctly created based on input values.

    Args:
        timeseries (TimeSeries): TimeSeries which assertion will be applied over.
        input_value (Any): Input value parameter during TimeSeries creation.
        input_unit (Any): Input unit parameter during TimeSeries creation.
        input_metadata (Any): Input metadata parameter during TimeSeries creation.
        input_timestamps (Any): Input timestamps parameter during TimeSeries creation.
        input_timestamps_unit (Any): Input timestamps unit parameter during TimeSeries creation.
        input_timestamps_metadata (Any): Input timestamps metadata parameter during TimeSeries creation.

    Raise:
        AssertionError: Raises if any assertion fails.
    """
    assert isinstance(timeseries, TimeSeries)
    assert_data_field_content(timeseries.value, input_value, input_unit, input_metadata)
    assert_data_field_content(timeseries.timestamps, input_timestamps, input_timestamps_unit, input_timestamps_metadata)


def assert_axis_content(
    axis: Axis,
    input_value: Any,
    input_unit: Any,
    input_metadata: Any,
    input_transformation_type: TransformationType,
    input_timestamps: Any = None,
    input_timestamps_unit: Any = None,
    input_timestamps_metadata: Any = None,
):
    """Assert that Axis was correctly created based on input values.

    Args:
        axis (Axis): Axis which assertion will be applied over.
        input_value (Any): Input value parameter during Axis creation.
        input_unit (Any): Input unit parameter during Axis creation.
        input_metadata (Any): Input metadata parameter during Axis creation.
        input_transformation_type(TransformationType): Input transformation type parameter during Axis creation.
        input_timestamps (Optional, Any): Input timestamps param during Axis creation. Defaults to None.
        input_timestamps_unit (Optional, Any): Input timestamps unit param during Axis creation. Defaults to None.
        input_timestamps_metadata (Optional, Any): Input timestamps metadata param on Axis creation. Defaults to None.

    Raise:
        AssertionError: Raises if any assertion fails.
    """
    assert isinstance(axis, Axis)
    assert axis.transformation_type == input_transformation_type
    if type(axis.value) is DataField:
        assert_data_field_content(axis.value, input_value, input_unit, input_metadata)
    elif type(axis.value) is TimeSeries:
        assert_timeseries_content(
            axis.value,
            input_value,
            input_unit,
            input_metadata,
            input_timestamps,
            input_timestamps_unit,
            input_timestamps_metadata,
        )
    else:
        raise AssertionError(f"{type(axis.value)} is a invalid type for Axis value property!")


def assert_data_fields_equality(df1: DataField, df2: DataField):
    """Assert two Assonant DataField are equal by comparing their content.

    Args:
        df1 (DataField): DataField which content will be compared to df2.
        df2 (DataField): DataField which content will be compared to df1.
    """
    assert_data_field_content(df1, df2.value, df2.unit, df2.extra_metadata)


def assert_timeseries_equality(ts1: TimeSeries, ts2: TimeSeries):
    """Assert two Assonant TimeSeries are equal by comparing their content.

    Args:
        ts1 (TimeSeries): TimeSeries which content will be compared to ts2.
        ts2 (TimeSeries): TimeSeries which content will be compared to ts1.
    """
    assert_timeseries_content(
        ts1,
        ts2.value.value,
        ts2.value.unit,
        ts2.value.extra_metadata,
        ts2.timestamps.value,
        ts2.value.unit,
        ts2.value.extra_metadata,
    )


def assert_axis_equality(ax1: Axis, ax2: Axis):
    """Assert two Assonant Axis are equal by comparing their content.

    Args:
        ax1 (Axis): Axis which content will be compared to ax2.
        ax2 (Axis): Axis which content will be compared to ax1.
    """

    # If Axis internal DataHandler is not from the same type, the assertion automatically fails
    assert type(ax1.value) is type(ax2.value)

    if type(ax1.value) is DataField:
        assert_axis_content(ax1, ax2.value.value, ax2.value.unit, ax2.value.extra_metadata, ax2.transformation_type)
    elif type(ax1.value) is TimeSeries:
        assert_axis_content(
            ax1,
            ax2.value.value.value,
            ax2.value.value.unit,
            ax2.value.value.extra_metadata,
            ax2.transformation_type,
            ax2.value.timestamps.value,
            ax2.value.timestamps.unit,
            ax2.value.timestamps.extra_metadata,
        )


def assert_component_equality(c1: Component, c2: Component):
    """Assert two Assonant Components are equal by comparing recursively their content dicts.

    Args:
        c1 (Component): Component which content will be compared to c2.
        c2 (Component): Component which content will be compared to c1.
    """

    # If both Components subcomponents keys are not equal, they are clearly different
    assert c1.subcomponents.keys() == c2.subcomponents.keys()

    # Check if Components fields and positions are equal
    assert c1.name == c2.name
    assert compare_dicts(c1.fields, c2.fields)
    assert compare_dicts(c1.positions, c2.positions)

    # Make a Recursive DFS to verify equality of all subcomponents
    for key in c1.subcomponents.keys():
        assert_component_equality(c1.subcomponents[key], c2.subcomponents[key])



"""Assonant Bluesky private module enums.

This submodule defines Enumerations classes used to standardize options related to bluesky tools.
"""

from .event_stream_name import EventStreamName

__all__ = ["EventStreamName"]



"""Assonant Detector component data class."""

from .component import Component


class Detector(Component):
    """Data class to handle all data required to define a detector."""


class DetectorModule(Component):
    """Data class to handle all data required to define a Module from a Detector."""


class DetectorROI(Component):
    """Data class to handle all data required to define a Region of Interest(ROI) taken from a Detector."""


class DetectorChannel(Component):
    """Data class to handle all data required to define a Channel from a Detector."""



"""Assonant metadata retriever.

This submodule defines Metadata Retrievers objects to allow retrieving important
metadata for Assonant from different sources.
"""

from .assonant_metadata_retriever import AssonantMetadataRetriever

__all__ = [
    "AssonantMetadataRetriever",
]



"""Assonant data retriever exceptions.

This submodule defines Exceptions used all over the data retriever module and its submodules.
"""

from .data_retrievers_exceptions import AssonantDataRetrieverError

__all__ = ["AssonantDataRetrieverError"]



import random
from typing import Generator, List, Union

import numpy as np
import pytest

from assonant.data_classes.components import Component
from assonant.data_classes.enums import TransformationType
from assonant.data_classes.factories import (
    AssonantComponentFactory,
    AssonantDataHandlerFactory,
    AssonantEntryFactory,
)
from assonant.naming_standards import BeamlineName


@pytest.fixture(scope="session")
def fixed_seed_random_generator() -> Generator:
    """
    Provides a NumPy random number generator with a fixed seed (42) to ensure reproducibility across tests.
    """
    return np.random.default_rng(seed=42)


# ======= Non-array data =======


@pytest.fixture(scope="session")
def example_int_value() -> int:
    """
    Provides a sample integer value for basic test cases.
    """
    return 1234


@pytest.fixture(scope="session")
def example_float_value() -> float:
    """
    Provides a sample float value for basic test cases.
    """
    return 43.21


@pytest.fixture(scope="session")
def example_string_value() -> float:
    """
    Provides a sample string value for basic test cases.
    """
    return "example value"


# ======= 1D data as numpy arrays =======


@pytest.fixture(scope="session")
def example_int_1d_array(fixed_seed_random_generator) -> np.typing.NDArray:
    """
    Provides a 1D NumPy array of integers (length 10) with values from 0 to 255.
    """
    return fixed_seed_random_generator.integers(0, 256, 10)


@pytest.fixture(scope="session")
def example_float_1d_array(fixed_seed_random_generator) -> np.typing.NDArray:
    """
    Provides a 1D NumPy array of floats (length 10) with values in [0, 1).
    """
    return fixed_seed_random_generator.random(10)


@pytest.fixture(scope="session")
def example_string_1d_array() -> np.typing.NDArray:
    """
    Provides a 1D NumPy array of sample string values.
    """
    return np.array(["a", "b", "c", "d", "e"])


@pytest.fixture(scope="session")
def example_float_1d_array_with_nan(example_float_1d_array) -> np.typing.NDArray:
    """
    Provides a copy of a float 1D NumPy array with NaNs injected at the start, middle, and end.
    Useful for testing NaN-safe operations.
    """
    float_1d_array_with_nan = example_float_1d_array.copy()
    float_1d_array_with_nan[0] = np.nan
    float_1d_array_with_nan[len(float_1d_array_with_nan) // 2] = np.nan
    float_1d_array_with_nan[-1] = np.nan
    return float_1d_array_with_nan


# ======= 2D data as numpy arrays =======


@pytest.fixture(scope="session")
def example_int_2d_array(fixed_seed_random_generator) -> np.typing.NDArray:
    """
    Provides a 10x10 NumPy array of random integers in [0, 256).
    """
    return fixed_seed_random_generator.integers(0, 256, (10, 10))


@pytest.fixture(scope="session")
def example_float_2d_array(fixed_seed_random_generator) -> np.typing.NDArray:
    """
    Provides a 10x10 NumPy array of random floats in [0, 1).
    """
    return fixed_seed_random_generator.random((10, 10))


@pytest.fixture(scope="session")
def example_string_2d_array() -> np.typing.NDArray:
    """
    Provides a 2D NumPy array (4x5) of sample strings.
    """
    return np.array(
        [["a", "b", "c", "d", "e"], ["f", "g", "h", "i", "j"], ["k", "l", "m", "n", "o"], ["p", "q", "r", "s", "t"]]
    )


@pytest.fixture(scope="session")
def example_float_2d_array_with_nan(example_float_2d_array) -> np.typing.NDArray:
    """
    Provides a copy of a float 2D NumPy array with NaNs injected at the top-left, center, and bottom-right.
    Useful for testing NaN handling in 2D data.
    """
    float_2d_array_with_nan = example_float_2d_array.copy()
    float_2d_array_with_nan[(0, 0)] = np.nan
    float_2d_array_with_nan[(len(float_2d_array_with_nan) // 2, len(float_2d_array_with_nan) // 2)] = np.nan
    float_2d_array_with_nan[(-1, -1)] = np.nan
    return float_2d_array_with_nan


# ======= 1D data as python list =======


@pytest.fixture(scope="session")
def example_int_1d_list() -> List[int]:
    """
    Provides a 1D list of 10 random integers in [0, 255] with fixed seed.
    """
    random.seed(42)
    return [random.randint(0, 255) for _ in range(10)]


@pytest.fixture(scope="session")
def example_float_1d_list() -> List[float]:
    """
    Provides a 1D list of 10 random float values in [0, 1) with fixed seed.
    """
    random.seed(42)
    return [random.random() for _ in range(10)]


@pytest.fixture(scope="session")
def example_string_1d_list() -> List[str]:
    """
    Provides a 1D list of sample string values.
    """
    return ["a", "b", "c", "d", "e"]


@pytest.fixture(scope="session")
def example_int_1d_list_with_none(example_int_1d_list) -> List[Union[int, None]]:
    """
    Provides a copy of the int list with None values at the start, middle, and end.
    """
    int_1d_list_with_none = example_int_1d_list.copy()
    int_1d_list_with_none[0] = None
    int_1d_list_with_none[len(int_1d_list_with_none) // 2] = None
    int_1d_list_with_none[-1] = None
    return int_1d_list_with_none


@pytest.fixture(scope="session")
def example_float_1d_list_with_none(example_float_1d_list) -> List[Union[float, None]]:
    """
    Provides a copy of the float list with None values at the start, middle, and end.
    """
    float_1d_list_with_none = example_float_1d_list.copy()
    float_1d_list_with_none[0] = None
    float_1d_list_with_none[len(float_1d_list_with_none) // 2] = None
    float_1d_list_with_none[-1] = None
    return float_1d_list_with_none


@pytest.fixture(scope="session")
def example_string_1d_list_with_none(example_string_1d_list) -> List[Union[str, None]]:
    """
    Provides a copy of the string list with None values at the start, middle, and end.
    """
    string_1d_list_with_none = example_string_1d_list.copy()
    string_1d_list_with_none[0] = None
    string_1d_list_with_none[len(string_1d_list_with_none) // 2] = None
    string_1d_list_with_none[-1] = None
    return string_1d_list_with_none


# ======= 2D data as python list =======


@pytest.fixture(scope="session")
def example_int_2d_list() -> List[List[int]]:
    """
    Provides a 2D list (10x10) of random integers in [0, 255] with fixed seed.
    """
    random.seed(42)
    return [[random.randint(0, 255) for _ in range(10)] for _ in range(10)]


@pytest.fixture(scope="session")
def example_float_2d_list() -> List[List[float]]:
    """
    Provides a 2D list (10x10) of random float values in [0, 1) with fixed seed.
    """
    random.seed(42)
    return [[random.random() for _ in range(10)] for _ in range(10)]


@pytest.fixture(scope="session")
def example_string_2d_list() -> List[List[str]]:
    """
    Provides a 2D list of sample string values (4x5 grid).
    """
    return [["a", "b", "c", "d", "e"], ["f", "g", "h", "i", "j"], ["k", "l", "m", "n", "o"], ["p", "q", "r", "s", "t"]]


@pytest.fixture(scope="session")
def example_int_2d_list_with_none(example_int_2d_list) -> List[List[Union[int, None]]]:
    """
    Provides a copy of the 2D int list with None values injected at the top-left, center, and bottom-right.
    """
    int_2d_list_with_none = example_int_2d_list.copy()
    int_2d_list_with_none[0][0] = None
    int_2d_list_with_none[len(int_2d_list_with_none) // 2][len(int_2d_list_with_none) // 2] = None
    int_2d_list_with_none[-1][-1] = None
    return int_2d_list_with_none


@pytest.fixture(scope="session")
def example_float_2d_list_with_none(example_float_2d_list) -> List[List[Union[float, None]]]:
    """
    Provides a copy of the 2D float list with None values injected at the top-left, center, and bottom-right.
    """
    float_2d_list_with_none = example_float_2d_list.copy()
    float_2d_list_with_none[0][0] = None
    float_2d_list_with_none[len(example_float_2d_list) // 2][len(example_float_2d_list) // 2] = None
    float_2d_list_with_none[-1][-1] = None
    return float_2d_list_with_none


@pytest.fixture(scope="session")
def example_string_2d_list_with_none(example_string_2d_list) -> List[List[Union[str, None]]]:
    """
    Provides a copy of the 2D string list with None values injected at the top-left, center, and bottom-right.
    """
    string_2d_list_with_none = example_string_2d_list.copy()
    string_2d_list_with_none[0][0] = None
    string_2d_list_with_none[len(string_2d_list_with_none) // 2][len(string_2d_list_with_none) // 2] = None
    string_2d_list_with_none[-1][-1] = None
    return string_2d_list_with_none


# ======= Additional metadata =======


@pytest.fixture(scope="session")
def example_metadata(fixed_seed_random_generator):
    """
    Provides a sample metadata dictionary with mixed types, including NumPy arrays.
    Useful for testing metadata compatibility and serialization.
    """
    return {
        "str metadata": "I'm a string metadata",
        "float metadata": 1.5,
        "int metadata": 42,
        "list metadata": [1, 2, 3, 4, 5],
        "numpy array metadata": fixed_seed_random_generator.integers(0, 10, 5),
    }


@pytest.fixture(scope="session")
def example_unit():
    """
    Provides a sample unit string representing a measurement unit.
    """
    return "Some measurement unit"


# ======= Assonant Entries =======


@pytest.fixture(scope="function")
def example_entry():
    return AssonantEntryFactory.create_entry(beamline_name=BeamlineName.CARNAUBA, name="entry")


# ======= Assonant Components =======
# Allow abstracting creation.
# Useful when testing different data types is not needed.


@pytest.fixture(scope="function")
def example_base_component():
    return AssonantComponentFactory.create_component_by_class_type(Component, "main_component")


@pytest.fixture(scope="function")
def example_subcomponent():
    return AssonantComponentFactory.create_component_by_class_type(Component, "subcomponent")


@pytest.fixture(scope="function")
def example_translation_axis(example_float_1d_array, example_unit, example_metadata):
    return AssonantDataHandlerFactory.create_axis(
        transformation_type=TransformationType.TRANSLATION,
        value=example_float_1d_array,
        unit=example_unit,
        extra_metadata=example_metadata,
    )


@pytest.fixture(scope="function")
def example_rotation_axis(example_float_1d_array, example_unit, example_metadata):
    return AssonantDataHandlerFactory.create_axis(
        transformation_type=TransformationType.ROTATION,
        value=example_float_1d_array,
        unit=example_unit,
        extra_metadata=example_metadata,
    )


@pytest.fixture(scope="function")
def example_int_1d_array_data_field(example_int_1d_array, example_unit, example_metadata):
    return AssonantDataHandlerFactory.create_data_field(
        value=example_int_1d_array, unit=example_unit, extra_metadata=example_metadata
    )


@pytest.fixture(scope="function")
def example_float_2d_array_data_field(example_float_2d_array, example_unit, example_metadata):
    return AssonantDataHandlerFactory.create_data_field(
        value=example_float_2d_array, unit=example_unit, extra_metadata=example_metadata
    )


@pytest.fixture(scope="function")
def example_string_data_field(example_string_value, example_unit, example_metadata):
    return AssonantDataHandlerFactory.create_data_field(
        value=example_string_1d_array, unit=example_unit, extra_metadata=example_metadata
    )



from typing import Any, Dict, List

import numpy as np
import pandas as pd

from assonant.data_classes.enums import ValuePlaceholders
from assonant.naming_standards import AcquisitionMoment

from ._assonant_metadata_retriever_interface import IAssonantMetadataRetriever
from .enums import AcquisitionType, CSVColumn
from .exceptions import AssonantMetadataRetrieverError


class CSVMetadataRetriever(IAssonantMetadataRetriever):
    """CSV Metadata Retriever."""

    def __init__(self, csv_file_path: str):

        self.invalid_values = [None, np.nan, ""]

        self.df = pd.read_csv(csv_file_path, sep=",", header=1)

        self._clean_data()

        self._pre_process()

    def _clean_data(self):
        """Apply data cleaning steps over dataframe. This methods should focus on steps which will make some
        kind of row/column removal/drop for some reason (e.g: Rows with not accept values).

        Raises:
            AssonantMetadataRetrieverError: Raised when a problem occurs in any of the data cleaning steps
        """
        try:
            # Remove all rows related to Experiment. Temporary limitation for tests!
            self.df = self.df[self.df[CSVColumn.CLASS.value] != "Experiment"]
        except Exception as e:
            raise AssonantMetadataRetrieverError(
                "[CLEAN-METADATA] CSV metadata retriever failed while filtering columns!"
            ) from e
        # pass # No data cleaning currently needed

    def _pre_process(self):
        """Apply pre-processing steps over dataframe. This methods should focus on steps which will make some kind
        of transformation over metadata from the Dataframe (e.g: Transform field values into str type).

        Raises:
            AssonantMetadataRetrieverError: Raised when a problem occurs in any of the pre-processing steps
        """
        try:
            # Replace np.nan value to None
            self.df = self.df.replace(np.nan, None)
        except Exception as e:
            raise AssonantMetadataRetrieverError(
                "[PRE-PROCESS] CSV metadata retriever failed while converting np.nan values to None!"
            ) from e
        # Transform Name and Class column on lower case
        try:
            # Transform all 'Name' strings into lowercase to avoid type error from users
            self.df[CSVColumn.NAME.value] = self.df[CSVColumn.NAME.value].apply(lambda x: x.lower())
        except Exception as e:
            raise AssonantMetadataRetrieverError(
                f"[PRE-PROCESS] CSV metadata retriever failed while lowering {CSVColumn.NAME.value} column strings! Cause may be a row with a missing value!"
            ) from e
        try:
            # Transform all 'Subcomponent of' strings into lowercase to correctly match 'Name' that
            # has just been lowercased.
            self.df[CSVColumn.SUBCOMPONENT_OF.value] = self.df[CSVColumn.SUBCOMPONENT_OF.value].apply(
                lambda x: x.lower() if x not in self.invalid_values else x
            )
        except Exception as e:
            raise AssonantMetadataRetrieverError(
                f"[PRE-PROCESS] CSV metadata retriever failed while lowering {CSVColumn.SUBCOMPONENT_OF.value} column strings!"
            ) from e
        try:
            # Remove empty space between names to match AssonantDataClass names
            self.df[CSVColumn.CLASS.value] = self.df[CSVColumn.CLASS.value].apply(lambda x: x.replace(" ", ""))
        except Exception as e:
            raise AssonantMetadataRetrieverError(
                f"[PRE-PROCESS] CSV metadata retriever failed while concatenating {CSVColumn.CLASS.value} column strings! Cause may be a row with a missing value!"
            ) from e
        # try:
        #    # Convert Experiment class to Entry class
        #    self.df[CSVColumn.CLASS.value] = self.df[CSVColumn.CLASS.value].replace('Experiment', 'Entry')
        # except Exception as e:
        #    raise AssonantMetadataRetrieverError(
        #        "[PRE-PROCESS] CSV metadata retriever failed while converting np.nan values to None!"
        #    ) from e

    def _get_row_by_pv_name(self, pv_name: str) -> pd.DataFrame:
        """Return DataFrame row that matches the passed PV name.

        Args:
            pv_name (str): PV name that identified the target row.

        Returns:
            pd.DataFrame: DataFrame row that matches the passed pv_name.
        """
        # Find row that matches the PV name and return a copy of it
        return self.df[self.df[CSVColumn.PV_NAME.value] == pv_name].copy(deep=True)

    def _convert_acquisition_moment_into_acquisition_type(
        self, acquisition_moment: AcquisitionMoment
    ) -> AcquisitionType:
        """Convert AcquisitionMoment value to its respective AcquisitionType equivalent

        Args:
            acquisition_moment (AcquisitionMoment): AcquisitionMoment Enum object to be converted

        Raises:
            AssonantMetadataRetrieverError: Raised if there is no valid convertion for passed AcquisitionMoment.

        Returns:
            AcquisitionType: AcquisitionType Enum object respective to the passed Acquisition Moment Enum object.
        """

        mapping = {
            AcquisitionMoment.START.value: AcquisitionType.SNAPSHOT,
            AcquisitionMoment.END.value: AcquisitionType.SNAPSHOT,
        }
        try:
            converted_value = mapping[acquisition_moment.value]
        except Exception as e:
            raise AssonantMetadataRetrieverError(
                f"There is no valid AcquisitionType convertion value for '{acquisition_moment.value}' AcquisitionMoment"
            ) from e
        return converted_value

    def get_pv_names_by_acquisition_moment(self, acquisition_moment: AcquisitionMoment) -> List[str]:
        """Return a List with all PV names related to the passed AcquisitionMoment.

        PS: Check IAssonantMetadataRetriever Interface for the return structure definition!

        Args:
            acquisition_type (AcquisitionMoment): Target Acquisition Moment used for selecting which
            PV names will be returned.

        Returns:
            List[str]: List containing all PV names related to passed AcquisitionMoment.
        """
        acquisition_type = self._convert_acquisition_moment_into_acquisition_type(acquisition_moment)
        filtered_df = self.df[self.df[CSVColumn.ACQUISITION_TYPE.value] == acquisition_type.value]
        filtered_df = filtered_df = filtered_df[
            ~filtered_df[CSVColumn.PV_NAME.value].isin([""]) & filtered_df[CSVColumn.PV_NAME.value].notna()
        ]

        # Iterate over rows and append the PV name to the result List
        result = [row[CSVColumn.PV_NAME.value] for _, row in filtered_df.iterrows()]
        return result

    def get_pvs_info(self, pv_names: List[str]) -> Dict[str, Dict[str, Any]]:
        """Return all information about the PV paired by the its name.

        PS: Check IAssonantMetadataRetriever Interface for the return structure definition!

        Args:
            pv_names (List[str]): List with PV names which info will be fetched.

        Returns:
            Dict[str, Dict[str, Any]]: Dictionary containing all PV metadata paired by the PV name.
        """

        filtered_df = self.df[self.df[CSVColumn.PV_NAME.value].isin(pv_names)]

        result = {
            row[CSVColumn.PV_NAME.value]: {
                "component_info": {
                    "name": row[CSVColumn.NAME.value],
                    "class": row[CSVColumn.CLASS.value],
                    **(
                        {"subcomponent_of": row[CSVColumn.SUBCOMPONENT_OF.value]}
                        if row[CSVColumn.SUBCOMPONENT_OF.value] not in self.invalid_values
                        else {}
                    ),
                },
                "data_handler_info": {
                    "name": row[CSVColumn.NEXUS_FIELD_NAME.value],
                    "value": ValuePlaceholders.VALUE_NOT_SET.value,
                    **(
                        {"unit": row[CSVColumn.UNIT_OF_MEASUREMENT.value]}
                        if row[CSVColumn.UNIT_OF_MEASUREMENT.value] not in self.invalid_values
                        else {}
                    ),
                    **(
                        {"transformation_type": row[CSVColumn.TRANSFORMATION_TYPE.value]}
                        if row[CSVColumn.TRANSFORMATION_TYPE.value] not in self.invalid_values
                        else {}
                    ),
                    "pv_name": row[CSVColumn.PV_NAME.value],
                },
            }
            for _, row in filtered_df.iterrows()
        }

        return result

    def get_subcomponents_mapping(self) -> Dict[str, List[str]]:
        """Return a Dictionary mapping component names with the name of its subcomponents.

        PS: Check IAssonantMetadataRetriever Interface for the return structure definition!

        Returns:
            Dict[str, List[str]]: Dictionary mapping the name of each Component that contains at
            least one Subcomponent with their respective List of Subcomponents names.
        """

        filtered_df = self.df[[CSVColumn.NAME.value, CSVColumn.SUBCOMPONENT_OF.value]].dropna()

        mapping = {}

        for component_name, subcomponent_names in filtered_df.groupby(CSVColumn.SUBCOMPONENT_OF.value)[
            CSVColumn.NAME.value
        ]:
            mapping[component_name] = list(set((subcomponent_names)))

        return mapping

    def get_component_info(self, component_name: str) -> Dict[str, Dict[str, Any]]:
        """Retrieve info about a specific Component based on passed component_name.

        PS: Check IAssonantMetadataRetriever Interface for the return structure definition!

        Args:
            component_name (str): Component name to retrieve information about.

        Returns:
            Dict[str, Dict[str, Any]]: Component info dictionary following the same subdictionary
            returned by get_pvs_info() method to represent component info.
        """
        query_result = self.df[self.df[CSVColumn.NAME.value] == component_name].head(1).iloc[0]

        if query_result.empty:
            return None
        else:
            component_info = {
                "component_info": {
                    "name": query_result[CSVColumn.NAME.value],
                    "class": query_result[CSVColumn.CLASS.value],
                    **(
                        {"subcomponent_of": query_result[CSVColumn.SUBCOMPONENT_OF.value]}
                        if query_result[CSVColumn.SUBCOMPONENT_OF.value] not in self.invalid_values
                        else {}
                    ),
                }
            }
        return component_info



"""Tests focused on validating Assonant methods from Data Classes submodule"""

import numpy as np
import pytest

from assonant.data_classes import AssonantEntryFactory, Entry
from assonant.data_classes.components import Beamline
from assonant.data_classes.exceptions import AssonantEntryFactoryError
from assonant.naming_standards import BeamlineName

# ================== Utilitary methods ==================


def assert_entry(entry: Entry, input_entry_name: str, input_beamline_name: BeamlineName):
    """Assert that Entry was correctly created based on input values

    Args:
        entry (Entry): Entry which assertion will be applied over.
        input_entry_name (str): Input entry_name parameter during Entry creation.
        input_beamline_name (BeamlineName): Input beamline_name parameter during Entry creation.

    Raise:
        AssertionError: Raises if any assertion fails.
    """
    assert isinstance(entry, Entry)
    assert entry.name == input_entry_name
    assert isinstance(entry.beamline, Beamline)
    assert entry.beamline.name == input_beamline_name.value


invalid_entry_names = [10, None, [1, 2, 3, 4], np.array([1, 2, 3, 4])]
invalid_beamline_names = ["name", 10, None, [1, 2, 3, 4], np.array([1, 2, 3, 4])]

# ================== Tests ==================


@pytest.mark.parametrize("entry_name, beamline_name", [("entry", bealine_name) for bealine_name in BeamlineName])
def test_entry_creation(entry_name, beamline_name):
    """
    Validate that AssonantEntryFactory correctly creates entries for all valid beamline names.

    Expected Behavior:
        - Given a valid `entry_name` and a valid `beamline_name` (from BeamlineName enum),
          the factory should:
            - Successfully instantiate an Assonant Entry object.
            - Set the entry's name and beamline attributes accordingly.
        - The resulting entry should pass all structural and semantic validations via `assert_entry`.
    """
    entry = AssonantEntryFactory.create_entry(entry_name=entry_name, beamline_name=beamline_name)

    assert_entry(entry, entry_name, beamline_name)


@pytest.mark.parametrize("entry_name", invalid_entry_names)
def test_entry_creation_invalid_name_type(entry_name):
    """
    Ensure that AssonantEntryFactory raises an error when provided with an invalid entry name.

    Expected Behavior:
        - For each invalid `entry_name` (e.g., non-string types, empty values),
          the factory should:
            - Raise an AssonantEntryFactoryError.
            - Not return an AssonantEntry instance.
        - This safeguards against improper usage and maintains entry name integrity.
    """
    with pytest.raises(AssonantEntryFactoryError):
        _ = AssonantEntryFactory.create_entry(entry_name=entry_name, beamline_name=BeamlineName.CARNAUBA)


@pytest.mark.parametrize("beamline_name", invalid_beamline_names)
def test_entry_creation_invalid_beamline_name_type(beamline_name):
    """
    Ensure that AssonantEntryFactory raises an error when provided with an invalid beamline name.

    Expected Behavior:
        - For each invalid `beamline_name` (Any values different from Assonant BeamlineName enum),
          the factory should:
            - Raise an AssonantEntryFactoryError.
            - Not return an AssonantEntry instance.
        - This safeguards against improper usage and maintains entry name integrity.
    """
    with pytest.raises(AssonantEntryFactoryError):
        _ = AssonantEntryFactory.create_entry(entry_name="entry", beamline_name=beamline_name)



"""Beamline names definitions."""

from enum import Enum


class BeamlineName(Enum):
    """Beamline Names."""

    ARIRANHA = "ARIRANHA"
    CARCARA = "CARCARA"
    CARNAUBA = "CARNAUBA"
    CATERETE = "CATERETE"
    CEDRO = "CEDRO"
    EMA = "EMA"
    IMBUIA = "IMBUIA"
    INGA = "INGA"
    IPE = "IPE"
    JATOBA = "JATOBA"
    HIBISCO = "HIBISCO"
    MANACA = "MANACA"
    MOGNO = "MOGNO"
    PAINEIRA = "PAINEIRA"
    PITANGA = "PITANGA"
    QUATI = "QUATI"
    QUIRIQUIRI = "QUIRIQUIRI"
    SABIA = "SABIA"
    SAGUI = "SAGUI"
    SAPE = "SAPE"
    SAPUCAIA = "SAPUCAIA"
    SERUENA = "SERUENA"
    SIBIPIRUNA = "SIBIPIRUNA"
    SUSSUARANA = "SUSSUARANA"
    TATU = "TATU"
    TEIU = "TEIU"
    TIMBO = "TIMBO"



"""Assonant Storage Ring component data class."""

from .component import Component


class StorageRing(Component):
    """Data class to handle all data required to define a particle accelerator storage ring."""



"""Assonant data classes exceptions.

This submodule defines Exceptions used all over the data classes module and its submodules.
"""

from .data_classes_exceptions import (
    AssonantDataClassesError,
    AxisCreationError,
    AxisInsertionError,
    DataHandlerInsertionError,
    DataHandlerTakeageError,
    FieldCreationError,
    FieldInsertionError,
    SubcomponentInsertionError,
)
from .factories_exceptions import (
    AssonantComponentFactoryError,
    AssonantDataHandlerFactoryError,
    AssonantEntryFactoryError,
)
from .hierarchizer_exceptions import AssonantHierarchizerError

__all__ = [
    "AssonantDataClassesError",
    "AssonantDataHandlerFactoryError",
    "AssonantComponentFactoryError",
    "AssonantEntryFactoryError",
    "AssonantHierarchizerError",
    "AxisCreationError",
    "AxisInsertionError",
    "DataHandlerInsertionError",
    "DataHandlerTakeageError",
    "FieldCreationError",
    "FieldInsertionError",
    "SubcomponentInsertionError",
]



"""Assonant Naming Standards.

Assonant Naming Standards, as the name suggest, defines naming standards to be used
inside and outside Assonant by application. The idea is to allow using the same
terminologies among applications and help in their interoperation and also
in the overall data standardization.
"""

from .acquisition_moment import AcquisitionMoment
from .beamline_name import BeamlineName
from .experiment_stage import ExperimentStage
from .experimental_technique import (
    ExperimentalTechniquesAcronyms,
    ExperimentalTechniquesFullNames,
)
from .facility_name import FacilityName

__all__ = [
    "AcquisitionMoment",
    "BeamlineName",
    "ExperimentStage",
    "ExperimentalTechniquesAcronyms",
    "ExperimentalTechniquesFullNames",
    "FacilityName",
]



from typing import Any, Dict, List, Union

from assonant._bluesky import DataParserFactory
from assonant._bluesky.enums import EventStreamName
from assonant.naming_standards import AcquisitionMoment

from ._assonant_data_retriever_interface import IAssonantDataRetriever
from .exceptions import AssonantDataRetrieverError


class BlueskyDataRetriever(IAssonantDataRetriever):
    """Bluesky Data Retriever."""

    def __init__(self, bluesky_data_source: Union[List[List[Union[str, Dict[str, Any]]]], str]):
        """Instantiates BlueskyDataRetriever to retrieve data from given data source.

        Args:
            bluesky_data_source (Union[List[List[Union[str, Dict[str, Any]]]], str]): DataSource containing
            the bluesky data. Currently, supported formats are: JSON file containing bluesky events or a
            List of bluesky documents data presented also as  a list. The list with the document data has
            2 positions, the 1st is the bluesky identification for the document
            (start, event, stop, descriptor, ...) and the 2nd position the dict containing the document
            data.

        Raises:
            AssonantDataRetrieverError: Failed to create the suitable Bluesky Data Parser for given data source.
        """
        try:
            self.bluesky_data_parser = DataParserFactory.create_data_parser(bluesky_data_source=bluesky_data_source)
        except Exception as e:
            raise AssonantDataRetrieverError(
                "BlueskyDataRetriever failed to create Bluesky DataParser for given data source."
            ) from e

    def _convert_acquisition_moment_into_event_stream_name(
        self, acquisition_moment: AcquisitionMoment
    ) -> EventStreamName:
        """Convert AcquisitionMoment value to its respective EventStreamName equivalent

        PS: On Bluesky, the existing DataStreams can be mapped to a respective AcquisitionMoment. Current mapping is:
            * Primary -> During
            * Monitor -> During
            * Baseline -> Start/End

        Args:
            acquisition_moment (AcquisitionMoment): AcquisitionMoment Enum object to be converted

        Raises:
            AssonantDataRetrieverError: Raised if there is no valid convertion for the passed AcquisitionMoment object.

        Returns:
            EventStreamName: EventStreamName Enum object respective to the passed Acquisition Moment Enum object.
        """

        mapping = {
            AcquisitionMoment.START.value: EventStreamName.BASELINE,
            AcquisitionMoment.END.value: EventStreamName.BASELINE,
            AcquisitionMoment.DURING.value: EventStreamName.PRIMARY,
        }
        try:
            converted_value = mapping[acquisition_moment.value]
        except Exception as e:
            raise AssonantDataRetrieverError(
                f"There is no valid EventStream convertion value for '{acquisition_moment.value}' AcquisitionMoment"
            ) from e
        return converted_value

    def get_pv_data_by_acquisition_moment(
        self, acquisition_moment: AcquisitionMoment
    ) -> Dict[str, Union[Any, List[Any]]]:
        """Return collected PV data related to specified AcquisitionMoment.

        The returned value is a dictionary as follow:

        {
            PV_NAME_1: [PV_VALUE_1, PV_VALUE_2, ...],
            PV_NAME_2: PV_VALUE_1,
            ...
        }

        PS: PV values may be a List/array of value or a single value

        Args:
            acquisition_moment (AcquisitionMoment): Target Acquisition Moment used to select
            data from which Acquisition Moment will be retrieved.

        Returns:
            Dict[str, Union[Any, List[Any]]]: Dictionary containing retrieved PV data.
        """
        if acquisition_moment == AcquisitionMoment.DURING:
            print(
                "WARNING!! Data from 'monitor' stream is not being saved!! If this is a requirement, contact "
                + "Assonant developers to develop this feature!!!"
            )

        event_stream_name = self._convert_acquisition_moment_into_event_stream_name(acquisition_moment)
        retrieved_data = self.bluesky_data_parser.query_data_by_stream_name(event_stream_name, with_config=False)

        # Convert data representation to fit the expected return structure
        formatted_retrieved_data = {}

        for pv_name, pv_data_list in retrieved_data.items():
            # Each position of pv_data_list is a tuple -> 1st position is the acquired value and 2nd the timestamp
            formatted_retrieved_data[pv_name] = [pv_data[0] for pv_data in pv_data_list]

        if acquisition_moment == AcquisitionMoment.START:
            for pv_name in formatted_retrieved_data:
                # Experiment Start snapshot is returned in the first position of the list for
                formatted_retrieved_data[pv_name] = formatted_retrieved_data[pv_name][0]
        elif acquisition_moment == AcquisitionMoment.END:
            for pv_name in formatted_retrieved_data:
                # Experiment End snapshot is returned in the last position of the list
                formatted_retrieved_data[pv_name] = formatted_retrieved_data[pv_name][-1]

        return formatted_retrieved_data
